# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import config
from gravitino.exceptions.base import IllegalArgumentException

from gravitino.rest.rest_message import RESTResponse


@dataclass
class BaseResponse(RESTResponse):
    """Represents a base response for REST API calls."""

    _code: int = field(metadata=config(field_name="code"))

    def code(self) -> int:
        return self._code

    def validate(self):
        """Validates the response code.
        @throws IllegalArgumentException if code value is negative.
        """
        if self._code < 0:
            raise IllegalArgumentException("code must be >= 0")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from enum import Enum


class Name(Enum):
    """
    The root type name of this type, representing all data types supported.
    """

    BOOLEAN = "BOOLEAN"
    """ The boolean type. """

    BYTE = "BYTE"
    """ The byte type. """

    SHORT = "SHORT"
    """ The short type. """

    INTEGER = "INTEGER"
    """ The integer type. """

    LONG = "LONG"
    """ The long type. """

    FLOAT = "FLOAT"
    """ The float type. """

    DOUBLE = "DOUBLE"
    """ The double type. """

    DECIMAL = "DECIMAL"
    """ The decimal type. """

    DATE = "DATE"
    """ The date type. """

    TIME = "TIME"
    """ The time type. """

    TIMESTAMP = "TIMESTAMP"
    """ The timestamp type. """

    INTERVAL_YEAR = "INTERVAL_YEAR"
    """ The interval year type. """

    INTERVAL_DAY = "INTERVAL_DAY"
    """ The interval day type. """

    STRING = "STRING"
    """ The string type. """

    VARCHAR = "VARCHAR"
    """ The varchar type. """

    FIXEDCHAR = "FIXEDCHAR"
    """ The char type with fixed length. """

    UUID = "UUID"
    """ The UUID type. """

    FIXED = "FIXED"
    """ The binary type with fixed length. """

    BINARY = "BINARY"
    """ The binary type with variable length. The length is specified in the type itself. """

    STRUCT = "STRUCT"
    """
    The struct type.
    A struct type is a complex type that contains a set of named fields, each with a type, 
    and optionally a comment.
    """

    LIST = "LIST"
    """
    The list type.
    A list type is a complex type that contains a set of elements, each with the same type.
    """

    MAP = "MAP"
    """
    The map type.
    A map type is a complex type that contains a set of key-value pairs, each with a key type 
    and a value type.
    """

    UNION = "UNION"
    """
    The union type.
    A union type is a complex type that contains a set of types.
    """

    NULL = "NULL"
    """ The null type. A null type represents a value that is null. """

    UNPARSED = "UNPARSED"
    """ The unparsed type. An unparsed type represents an unresolvable type. """

    EXTERNAL = "EXTERNAL"
    """ The external type. An external type represents a type that is not supported. """


# Define the Type interface (abstract base class)
class Type(ABC):
    @abstractmethod
    def name(self) -> Name:
        """Returns the generic name of the type."""
        pass

    @abstractmethod
    def simple_string(self) -> str:
        """Returns a readable string representation of the type."""
        pass


# Define base classes
class PrimitiveType(Type, ABC):
    """Base class for all primitive types."""

    pass


class NumericType(PrimitiveType, ABC):
    """Base class for all numeric types."""

    pass


class DateTimeType(PrimitiveType, ABC):
    """Base class for all date/time types."""

    # Indicates that precision for the date/time type was not explicitly set by the user.
    # The value should be converted to the catalog's default precision.
    DATE_TIME_PRECISION_NOT_SET = -1

    # Represents the minimum precision range for timestamp, time and other date/time types.
    # The minimum precision is 0, which means second-level precision.
    MIN_ALLOWED_PRECISION = 0

    # Represents the maximum precision allowed for timestamp, time and other date/time types.
    # The maximum precision is 12, which means picosecond-level precision.
    MAX_ALLOWED_PRECISION = 12


class IntervalType(PrimitiveType, ABC):
    """Base class for all interval types."""

    pass


class ComplexType(Type, ABC):
    """Base class for all complex types, including struct, list, map, and union."""

    pass


class IntegralType(NumericType, ABC):
    """Base class for all integral types."""

    _signed: bool

    def __init__(self, signed: bool):
        self._signed = signed

    def signed(self) -> bool:
        """Returns True if the integer type is signed, False otherwise."""
        return self._signed


class FractionType(NumericType, ABC):
    """Base class for all fractional types."""

    pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import re
from typing import Dict

from gravitino.client.gravitino_client import GravitinoClient
from gravitino.auth.default_oauth2_token_provider import DefaultOAuth2TokenProvider
from gravitino.auth.oauth2_token_provider import OAuth2TokenProvider
from gravitino.auth.simple_auth_provider import SimpleAuthProvider
from gravitino.filesystem.gvfs_config import GVFSConfig
from gravitino.name_identifier import NameIdentifier
from gravitino.exceptions.base import GravitinoRuntimeException

_identifier_pattern = re.compile(
    "^(?:gvfs://)?fileset/([^/]+)/([^/]+)/([^/]+)(?:/[^/]+)*/?$"
)


def _check_auth_config(auth_type: str, config_key: str, config_value: str):
    """Check if the config value is null.
    :param auth_type: The auth type
    :param config_key: The config key
    :param config_value: The config value
    """
    if config_value is None:
        raise GravitinoRuntimeException(
            f"{config_key} should not be null"
            f" if {GVFSConfig.AUTH_TYPE} is set to {auth_type}."
        )


def create_client(
    options: Dict[str, str],
    server_uri: str,
    metalake_name: str,
    request_headers: dict = None,
    client_config: dict = None,
):
    """Create the Gravitino client.
    :param options: The options
    :param server_uri: The server URI
    :param metalake_name: The metalake name
    :param request_headers: The request headers
    :param client_config: The client config
    :return The Gravitino client
    """
    auth_type = (
        GVFSConfig.SIMPLE_AUTH_TYPE
        if options is None
        else options.get(GVFSConfig.AUTH_TYPE, GVFSConfig.SIMPLE_AUTH_TYPE)
    )

    if auth_type == GVFSConfig.SIMPLE_AUTH_TYPE:
        return GravitinoClient(
            uri=server_uri,
            metalake_name=metalake_name,
            auth_data_provider=SimpleAuthProvider(),
            request_headers=request_headers,
            client_config=client_config,
        )

    if auth_type == GVFSConfig.OAUTH2_AUTH_TYPE:
        oauth2_server_uri = options.get(GVFSConfig.OAUTH2_SERVER_URI)
        _check_auth_config(auth_type, GVFSConfig.OAUTH2_SERVER_URI, oauth2_server_uri)

        oauth2_credential = options.get(GVFSConfig.OAUTH2_CREDENTIAL)
        _check_auth_config(auth_type, GVFSConfig.OAUTH2_CREDENTIAL, oauth2_credential)

        oauth2_path = options.get(GVFSConfig.OAUTH2_PATH)
        _check_auth_config(auth_type, GVFSConfig.OAUTH2_PATH, oauth2_path)

        oauth2_scope = options.get(GVFSConfig.OAUTH2_SCOPE)
        _check_auth_config(auth_type, GVFSConfig.OAUTH2_SCOPE, oauth2_scope)

        oauth2_token_provider: OAuth2TokenProvider = DefaultOAuth2TokenProvider(
            oauth2_server_uri, oauth2_credential, oauth2_path, oauth2_scope
        )

        return GravitinoClient(
            uri=server_uri,
            metalake_name=metalake_name,
            auth_data_provider=oauth2_token_provider,
            request_headers=request_headers,
            client_config=client_config,
        )

    raise GravitinoRuntimeException(
        f"Authentication type {auth_type} is not supported."
    )


def extract_identifier(metalake_name: str, path: str):
    """Extract the fileset identifier from the path.
    :param metalake_name: The metalake name of the fileset.
    :param path: The virtual fileset path, format is
    [gvfs://fileset]/{fileset_catalog}/{fileset_schema}/{fileset_name}[/sub_path]
    :return The fileset identifier
    """
    if not metalake_name or metalake_name.isspace():
        raise ValueError("Virtual path cannot be null or empty.")

    if not path or (isinstance(path, str) and path.isspace()):
        raise GravitinoRuntimeException(
            "path which need be extracted cannot be null or empty."
        )

    match = _identifier_pattern.match(str(path))
    if match and len(match.groups()) == 3:
        return NameIdentifier.of(
            metalake_name, match.group(1), match.group(2), match.group(3)
        )
    raise GravitinoRuntimeException(
        f"path: `{path}` doesn't contains valid identifier."
    )


def get_sub_path_from_virtual_path(identifier: NameIdentifier, virtual_path: str):
    """Get the sub path from the virtual path.
    :param identifier: The identifier of the fileset.
    :param virtual_path: The virtual fileset path, format is
    fileset/{fileset_catalog}/{fileset_schema}/{fileset_name}[/sub_path]
    :return The sub path.
    """
    if not virtual_path or virtual_path.isspace():
        raise ValueError("Virtual path cannot be null or empty.")

    gvfs_path_prefix = "fileset/"
    if not virtual_path.startswith(gvfs_path_prefix):
        raise ValueError(f"Virtual path should start with '{gvfs_path_prefix}'")

    prefix = to_gvfs_path_prefix(identifier)
    if not virtual_path.startswith(prefix):
        raise ValueError(
            f"Virtual path '{virtual_path}' doesn't match fileset identifier '{identifier}'"
        )

    return virtual_path[len(prefix) :]


def to_gvfs_path_prefix(identifier: NameIdentifier):
    """Convert the fileset identifier to the virtual path prefix.
    :param identifier: The name identifier of the fileset.
    :return The virtual path prefix, format is: fileset/{fileset_catalog}/{fileset_schema}/{fileset_name}
    """
    if len(identifier.namespace().levels()) != 3:
        raise ValueError(
            f"The namespace of the identifier should have 3 levels, but got {len(identifier.namespace().levels())}"
        )

    if not identifier.name():
        raise ValueError("The identifier name should not be null or empty.")

    return (
        f"fileset/{identifier.namespace().level(1)}"
        f"/{identifier.namespace().level(2)}"
        f"/{identifier.name()}"
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Dict

from gravitino.api.rel.expressions.sorts.null_ordering import NullOrdering
from gravitino.api.rel.expressions.sorts.sort_direction import SortDirection
from gravitino.api.rel.types.json_serdes import JsonSerializable
from gravitino.dto.rel.expressions.json_serdes._helper.serdes_utils import SerdesUtils
from gravitino.dto.rel.sort_order_dto import SortOrderDTO
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class SortOrderSerdes(SerdesUtilsBase, JsonSerializable[SortOrderDTO]):
    """Custom JSON serializer/deserializer for SortOrderDTO objects."""

    @classmethod
    def serialize(cls, data_type: SortOrderDTO) -> dict[str, Any]:
        """
        Serialize the given data into a dictionary.

        Args:
            data_type (SortOrderDTO): The data to be serialized.

        Returns:
            dict[str, Any]: The serialized data.
        """

        return {
            cls.SORT_TERM: SerdesUtils.write_function_arg(data_type.sort_term()),
            cls.DIRECTION: str(data_type.direction()),
            cls.NULL_ORDERING: str(data_type.null_ordering()),
        }

    @classmethod
    def deserialize(cls, data: dict[str, Any]) -> SortOrderDTO:
        """
        Deserialize the given data into a `SortOrderDTO` instance.

        Args:
            data (dict[str, Any]): The data to be deserialized.

        Returns:
            SortOrderDTO: The deserialized `SortOrderDTO` instance.
        """

        Precondition.check_argument(
            isinstance(data, Dict) and len(data) > 0,
            f"Cannot parse sort order from invalid JSON: {data}",
        )
        Precondition.check_argument(
            cls.SORT_TERM in data,
            f"Cannot parse sort order from missing sort term: {data}",
        )
        sort_term_data = data[cls.SORT_TERM]
        Precondition.check_argument(
            sort_term_data is not None, "expression cannot be null"
        )
        direction_data = data.get(cls.DIRECTION)
        null_order_data = data.get(cls.NULL_ORDERING)
        sort_term = SerdesUtils.read_function_arg(data[cls.SORT_TERM])
        direction = (
            SortDirection.from_string(direction_data)
            if direction_data
            else SortDirection.ASCENDING
        )
        null_order = (
            NullOrdering(null_order_data)
            if null_order_data
            else direction.default_null_ordering()
        )

        return SortOrderDTO(sort_term, direction, null_order)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class GravitinoRuntimeException(RuntimeError):
    """Base class for all Gravitino runtime exceptions."""

    def __init__(self, message, *args):
        super().__init__(message % args)


class RESTException(RuntimeError):
    """An exception thrown when a REST request fails."""

    def __init__(self, message, *args):
        super().__init__(message % args)


class IllegalArgumentException(ValueError):
    """Base class for all exceptions thrown when arguments are invalid."""

    def __init__(self, message, *args):
        super().__init__(message % args)


class IllegalNameIdentifierException(IllegalArgumentException):
    """An exception thrown when a name identifier is invalid."""


class IllegalNamespaceException(IllegalArgumentException):
    """An exception thrown when a namespace is invalid."""


class InternalError(GravitinoRuntimeException):
    """Base class for all exceptions thrown internally"""


class NotFoundException(GravitinoRuntimeException):
    """Base class for all exceptions thrown when a resource is not found."""


class NoSuchSchemaException(NotFoundException):
    """An exception thrown when a schema is not found."""


class NoSuchFilesetException(NotFoundException):
    """Exception thrown when a file with specified name is not existed."""


class NoSuchLocationNameException(NotFoundException):
    """Exception thrown when a fileset location name is not found."""


class NoSuchCredentialException(NotFoundException):
    """Exception thrown when a credential with specified credential type is not existed."""


class NoSuchMetalakeException(NotFoundException):
    """An exception thrown when a metalake is not found."""


class NoSuchCatalogException(NotFoundException):
    """An exception thrown when a catalog is not found."""


class NoSuchModelException(NotFoundException):
    """An exception thrown when a model is not found."""


class NoSuchModelVersionException(NotFoundException):
    """An exception thrown when a model version is not found."""


class NoSuchModelVersionURINameException(NotFoundException):
    """An exception thrown when a URI name of a model version is not found."""


class AlreadyExistsException(GravitinoRuntimeException):
    """Base exception thrown when an entity or resource already exists."""


class MetalakeAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a metalake already exists."""


class SchemaAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a schema already exists."""


class CatalogAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a resource already exists."""


class ModelAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a model already exists."""


class ModelVersionAliasesAlreadyExistException(AlreadyExistsException):
    """An exception thrown when model version with aliases already exists."""


class NotEmptyException(GravitinoRuntimeException):
    """Base class for all exceptions thrown when a resource is not empty."""


class NonEmptySchemaException(NotEmptyException):
    """Exception thrown when a namespace is not empty."""


class InUseException(GravitinoRuntimeException):
    """Base class for all exceptions thrown when an entity is in use and cannot be deleted."""


class MetalakeInUseException(InUseException):
    """An exception thrown when a metalake is in use and cannot be deleted."""


class CatalogInUseException(InUseException):
    """An Exception thrown when a catalog is in use and cannot be deleted."""


class NotInUseException(GravitinoRuntimeException):
    """Base class for all exceptions thrown when an entity is not in use."""


class MetalakeNotInUseException(NotInUseException):
    """An exception thrown when operating on not in use metalake."""


class CatalogNotInUseException(NotInUseException):
    """An exception thrown when operating on not in use catalog."""


class UnsupportedOperationException(GravitinoRuntimeException):
    """Base class for all exceptions thrown when an operation is unsupported"""


class UnknownError(RuntimeError):
    """An exception thrown when other unknown exception is thrown"""


class ConnectionFailedException(GravitinoRuntimeException):
    """An exception thrown when connect to catalog failed."""


class UnauthorizedException(GravitinoRuntimeException):
    """An exception thrown when a user is not authorized to perform an action."""


class BadRequestException(GravitinoRuntimeException):
    """An exception thrown when the request is invalid."""


class IllegalStateException(GravitinoRuntimeException):
    """An exception thrown when the state is invalid."""


class NoSuchTagException(NotFoundException):
    """An exception thrown when a tag with specified name is not existed."""


class TagAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a tag with specified name already exists."""


class TagAlreadyAssociatedException(AlreadyExistsException):
    """Exception thrown when a tag with specified name already associated to a metadata object."""


class JobTemplateAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a job template with specified name already exists."""


class IllegalJobTemplateOperationException(IllegalArgumentException):
    """An exception thrown when an illegal operation is attempted on a job template.

    This exception is raised when attempting to modify or delete a built-in job template,
    or when trying to create a user template with a reserved name (e.g., starting with 'builtin-').
    """


class NoSuchJobTemplateException(NotFoundException):
    """An exception thrown when a job template with specified name is not found."""


class NoSuchJobException(NotFoundException):
    """An exception thrown when a job with specified name is not found."""


class ForbiddenException(GravitinoRuntimeException):
    """An exception thrown when a user is forbidden to perform an action."""


class NoSuchTableException(NotFoundException):
    """An exception thrown when a table with specified name is not existed."""


class NoSuchPartitionException(NotFoundException):
    """An exception thrown when a partition with specified name is not existed."""


class PartitionAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a partition with specified name already exists."""


class TableAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a table already exists."""


class NoSuchFunctionException(NotFoundException):
    """An exception thrown when a function with specified name is not found."""


class FunctionAlreadyExistsException(AlreadyExistsException):
    """An exception thrown when a function already exists."""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import ClassVar, Optional, Union, overload

from gravitino.api.metadata_object import MetadataObject
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.utils.precondition import Precondition


class MetadataObjects:
    """The helper class for `MetadataObject`."""

    METADATA_OBJECT_RESERVED_NAME: ClassVar[str] = "*"
    """The reserved name for the metadata object."""

    DOT_SPLITTER: ClassVar[str] = "."
    DOT_JOINER: ClassVar[str] = "."

    _NAMES_LEN_CONDS: ClassVar[dict[int, set[MetadataObject.Type]]] = {
        1: {
            MetadataObject.Type.METALAKE,
            MetadataObject.Type.CATALOG,
            MetadataObject.Type.ROLE,
            MetadataObject.Type.TAG,
            MetadataObject.Type.POLICY,
            MetadataObject.Type.JOB,
            MetadataObject.Type.JOB_TEMPLATE,
        },
        2: {MetadataObject.Type.SCHEMA},
        3: {
            MetadataObject.Type.FILESET,
            MetadataObject.Type.TABLE,
            MetadataObject.Type.TOPIC,
            MetadataObject.Type.MODEL,
        },
        4: {MetadataObject.Type.COLUMN},
    }

    @staticmethod
    @overload
    def of(name_or_names: list[str], type_: MetadataObject.Type) -> MetadataObject: ...

    @staticmethod
    @overload
    def of(
        name_or_names: str, type_: MetadataObject.Type, parent: Optional[str] = None
    ) -> MetadataObject: ...

    @staticmethod
    def of(
        name_or_names: Union[str, list[str]],
        type_: MetadataObject.Type,
        parent: Optional[str] = None,
    ) -> MetadataObject:
        if isinstance(name_or_names, str):
            name = name_or_names
            full_name = (
                MetadataObjects.DOT_JOINER.join([parent, name]) if parent else name
            )
            return MetadataObjects.parse(full_name, type_)
        names_len = len(name_or_names)
        Precondition.check_argument(
            names_len > 0,
            "Cannot create a metadata object with no names",
        )
        Precondition.check_argument(
            names_len <= 4,
            "Cannot create a metadata object with the name length which is greater than 4",
        )
        Precondition.check_argument(
            names_len != 1 or type_ in MetadataObjects._NAMES_LEN_CONDS[1],
            "If the length of names is 1, it must be the CATALOG, METALAKE, or ROLE type",
        )
        Precondition.check_argument(
            names_len != 2 or type_ in MetadataObjects._NAMES_LEN_CONDS[2],
            "If the length of names is 2, it must be the SCHEMA type",
        )
        Precondition.check_argument(
            names_len != 3 or type_ in MetadataObjects._NAMES_LEN_CONDS[3],
            "If the length of names is 3, it must be FILESET, TABLE, TOPIC or MODEL",
        )
        Precondition.check_argument(
            names_len != 4 or type_ in MetadataObjects._NAMES_LEN_CONDS[4],
            "If the length of names is 4, it must be COLUMN",
        )
        names = name_or_names
        for name in names:
            MetadataObjects.check_name(name)
        return MetadataObjects.MetadataObjectImpl(
            MetadataObjects.get_last_name(names),
            type_,
            MetadataObjects.get_parent_full_name(names),
        )

    @staticmethod
    def parent(object_: MetadataObject) -> Optional[MetadataObject]:
        """Get the parent metadata object of the given metadata object.

        Args:
            object_ (MetadataObject): The metadata object

        Returns:
            Optional[MetadataObject]:
                The parent metadata object if it exists, otherwise `None`
        """
        if object_ is None:
            return None

        object_type = object_.type()
        # Return None if the object is the root object
        if object_type in {
            MetadataObject.Type.METALAKE,
            MetadataObject.Type.CATALOG,
            MetadataObject.Type.ROLE,
        }:
            return None

        parent_type = None
        if object_type is MetadataObject.Type.COLUMN:
            parent_type = MetadataObject.Type.TABLE
        elif object_type in {
            MetadataObject.Type.TABLE,
            MetadataObject.Type.FILESET,
            MetadataObject.Type.TOPIC,
            MetadataObject.Type.MODEL,
        }:
            parent_type = MetadataObject.Type.SCHEMA
        elif object_type is MetadataObject.Type.SCHEMA:
            parent_type = MetadataObject.Type.CATALOG
        else:
            raise IllegalArgumentException(
                f"Unexpected to reach here for metadata object type: {object_type.value}"
            )

        return MetadataObjects.parse(object_.parent(), parent_type)

    @staticmethod
    def get_parent_full_name(names: list[str]) -> Optional[str]:
        """Get the parent full name of the given full name.

        Args:
            names (list[str]): The names of the metadata object

        Returns:
            Optional[str]: The parent full name if it exists, otherwise `None`.
        """
        names_len = len(names)
        if names_len <= 1:
            return None
        return MetadataObjects.DOT_JOINER.join(names[:-1])

    @staticmethod
    def get_last_name(names: list[str]) -> str:
        return names[-1]

    @staticmethod
    def check_name(name: str) -> None:
        Precondition.check_argument(
            name is not None,
            "Cannot create a metadata object with null name",
        )
        Precondition.check_argument(
            name != MetadataObjects.METADATA_OBJECT_RESERVED_NAME,
            "Cannot create a metadata object with `*` name.",
        )

    @staticmethod
    def parse(full_name: str, type_: MetadataObject.Type) -> MetadataObject:
        Precondition.check_argument(
            full_name is not None and len(full_name.strip()) > 0,
            "Metadata object full name cannot be blank",
        )

        parts = full_name.split(MetadataObjects.DOT_SPLITTER)
        if type_ is MetadataObject.Type.ROLE:
            return MetadataObjects.of([full_name], MetadataObject.Type.ROLE)

        return MetadataObjects.of(parts, type_)

    class MetadataObjectImpl(MetadataObject):
        def __init__(
            self, name: str, type_: MetadataObject.Type, parent: Optional[str] = None
        ) -> None:
            self._name = name
            self._type = type_
            self._parent = parent

        def name(self) -> str:
            return self._name

        def type(self) -> MetadataObject.Type:
            return self._type

        def parent(self) -> Optional[str]:
            return self._parent

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, MetadataObjects.MetadataObjectImpl):
                return False
            return (
                self._name == value.name()
                and self._type == value.type()
                and self._parent == value.parent()
            )

        def __hash__(self) -> int:
            return hash((self._name, self._parent, self._type))

        def __str__(self) -> str:
            return f"MetadataObject: [fullName={self.full_name()}], [type={self.type().value}]"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from enum import Enum
from typing import Optional, Dict

from gravitino.api.auditable import Auditable


class Fileset(Auditable):
    """An interface representing a fileset under a schema Namespace. A fileset is a virtual
    concept of the file or directory that is managed by Gravitino. Users can create a fileset object
    to manage the non-tabular data on the FS-like storage. The typical use case is to manage the
    training data for AI workloads. The major difference compare to the relational table is that the
    fileset is schema-free, the main property of the fileset is the storage location of the
    underlying data.

    Fileset defines the basic properties of a fileset object. A catalog implementation
    with FilesetCatalog should implement this interface.
    """

    PROPERTY_MULTIPLE_LOCATIONS_PREFIX = "location-"
    """The prefix of the location name in the property at the catalog/schema level."""

    PROPERTY_DEFAULT_LOCATION_NAME = "default-location-name"
    """The property name for the default location name of the fileset."""

    LOCATION_NAME_UNKNOWN = "unknown"
    """The reserved location name to indicate the location name is unknown."""

    PROPERTY_LOCATION_PLACEHOLDER_PREFIX = "placeholder-"
    """The prefix of fileset placeholder property"""

    PROPERTY_CATALOG_PLACEHOLDER = "placeholder-catalog"
    """The reserved property name for the catalog name placeholder, when creating a fileset, all
    placeholders as {{catalog}} will be replaced by the catalog name"""

    PROPERTY_SCHEMA_PLACEHOLDER = "placeholder-schema"
    """The reserved property name for the schema name placeholder, when creating a fileset, all
    placeholders as {{schema}} will be replaced by the schema name"""

    PROPERTY_FILESET_PLACEHOLDER = "placeholder-fileset"
    """The reserved property name for the fileset name placeholder, when creating a fileset, all
    placeholders as {{fileset}} will be replaced by the fileset name"""

    class Type(Enum):
        """An enum representing the type of the fileset object."""

        MANAGED = "managed"
        """Fileset is managed by Gravitino. 
        When specified, the data will be deleted when the fileset object is deleted"""

        EXTERNAL = "external"
        """Fileset is not managed by Gravitino. 
        When specified, the data will not be deleted when the fileset object is deleted"""

    @abstractmethod
    def name(self) -> str:
        """
        Returns:
            Name of the fileset object.
        """
        pass

    @abstractmethod
    def comment(self) -> Optional[str]:
        """
        Returns:
            The comment of the fileset object. Null is returned if no comment is set.
        """
        pass

    @abstractmethod
    def type(self) -> Type:
        """
        @Returns:
            The type of the fileset object.
        """
        pass

    def storage_location(self) -> str:
        """Get the unnamed storage location of the file or directory path managed by this fileset object.

        The returned storageLocation can be either the one specified when creating the fileset object
        (using storageLocation field or storageLocations field), or the one specified at the catalog/schema
        level (using property "location" or properties with prefix "location-") if the fileset object is
        created under this catalog/schema.

        The storageLocation at each level can contain placeholders, formatted as {{name}}, which will
        be replaced by the corresponding fileset property value when the fileset object is created.
        The placeholder property in the fileset object is formed as "placeholder-{{name}}". For example,
        if the storageLocation is "file:///path/{{schema}}-{{fileset}}-{{version}}", and the fileset
        object "catalog1.schema1.fileset1" has the property "placeholder-version" set to "v1",
        then the storageLocation will be "file:///path/schema1-fileset1-v1".

        For managed fileset, the storageLocation can be:

        1) The one specified when creating the fileset object, and the placeholders in the
           storageLocation will be replaced by the placeholder value specified in the fileset properties.

        2) When catalog property "location" is specified but schema property "location" is not specified,
           then the storageLocation will be:
            a. "{catalog location}/schemaName/filesetName" - if {catalog location} has no placeholders
            b. "{catalog location}" - placeholders in {catalog location} will be replaced by values
               specified in fileset properties

        3) When catalog property "location" is not specified but schema property "location" is specified,
           then the storageLocation will be:
            a. "{schema location}/filesetName" - if {schema location} has no placeholders
            b. "{schema location}" - placeholders in {schema location} will be replaced by values
               specified in fileset properties

        4) When both catalog property "location" and schema property "location" are specified,
           then the storageLocation will be:
            a. "{schema location}/filesetName" - if {schema location} has no placeholders
            b. "{schema location}" - placeholders in {schema location} will be replaced by values
               specified in fileset properties

        5) null value - when catalog property "location", schema property "location", storageLocation
        field of fileset, and "unknown" location in storageLocations are not specified.

        For external fileset, the storageLocation can be:
        1) The one specified when creating the fileset object, and the placeholders in the
           storageLocation will be replaced by the placeholder value specified in the fileset properties.

        Returns:
            str: The storage location of the fileset object.
        """
        return self.storage_locations().get(self.LOCATION_NAME_UNKNOWN)

    @abstractmethod
    def storage_locations(self) -> Dict[str, str]:
        """
        Get the storage location name and corresponding path of the file or directory path that is
        managed by this fileset object. The key is the name of the storage location and the value is
        the storage location path.

        Each storageLocation in the values can either be the one specified when creating the fileset
        object, or the one specified in the catalog / schema level if the fileset object is created
        under this catalog / schema.

        The "unknown" location name is reserved to indicate the default storage location of the fileset.
        It can be specified in catalog / schema level by the property "location" or in the fileset level
        by the field "storageLocation". Other location names can be specified in the fileset level by the
        key-value pairs in the field "storageLocations", and by "location-{name}" properties in the
        catalog / schema level.

        The storageLocation at each level can contain placeholders, formatted as {{name}}, which will
        be replaced by the corresponding fileset property value when the fileset object is created.
        The placeholder property in the fileset object is formed as "placeholder-{{name}}". For example,
        if the storageLocation is "file:///path/{{schema}}-{{fileset}}-{{version}}", and the fileset
        object "catalog1.schema1.fileset1" has the property "placeholder-version" set to "v1",
        then the storageLocation will be "file:///path/schema1-fileset1-v1".

        For managed fileset, the storageLocation can be:

        1) The one specified when creating the fileset object, and the placeholders in the
           storageLocation will be replaced by the placeholder value specified in the fileset properties.

        2) When catalog property "location" is specified but schema property "location" is not specified,
           then the storageLocation will be:
            a. "{catalog location}/schemaName/filesetName" - if {catalog location} has no placeholders
            b. "{catalog location}" - placeholders in {catalog location} will be replaced by values
               specified in fileset properties

        3) When catalog property "location" is not specified but schema property "location" is specified,
           then the storageLocation will be:
            a. "{schema location}/filesetName" - if {schema location} has no placeholders
            b. "{schema location}" - placeholders in {schema location} will be replaced by values
               specified in fileset properties

        4) When both catalog property "location" and schema property "location" are specified,
           then the storageLocation will be:
            a. "{schema location}/filesetName" - if {schema location} has no placeholders
            b. "{schema location}" - placeholders in {schema location} will be replaced by values
               specified in fileset properties

        5) When there is no location specified in catalog level, schema level, storageLocation of fileset,
           and storageLocations of fileset at the same time, this situation is illegal.

        For external fileset, the storageLocation can be:
        1) The one specified when creating the fileset object, and the placeholders in the
           storageLocation will be replaced by the placeholder value specified in the fileset properties.

        :return: The storage locations of the fileset object, the key is the name of the storage
        location and the value is the storage location path.
        """
        pass

    @abstractmethod
    def properties(self) -> Dict[str, str]:
        """
        Returns:
            The properties of the fileset object. Empty map is returned if no properties are set.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.dto.metadata_object_dto import MetadataObjectDTO
from gravitino.dto.responses.base_response import BaseResponse
from gravitino.utils.precondition import Precondition


@dataclass
class MetadataObjectListResponse(BaseResponse):
    """Represents a response containing a list of metadata objects."""

    _metadata_objects: list[MetadataObjectDTO] = field(
        default_factory=list, metadata=config(field_name="metadataObjects")
    )

    def metadata_objects(self) -> list[MetadataObjectDTO]:
        """
        Retrieve the list of metadata objects.

        Returns:
            list[MetadataObjectDTO]: The list of metadata objects.
        """
        return self._metadata_objects

    def validate(self) -> None:
        """
        Validates the response data.
        """
        super().validate()

        for metadata_object in self._metadata_objects:
            Precondition.check_argument(
                metadata_object is not None
                and metadata_object.type() is not None
                and (name := metadata_object.name()) is not None
                and name.strip() != "",
                "metadataObject must not be null and its field cannot null or empty",
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from abc import abstractmethod
from enum import Enum, unique

from gravitino.api.rel.partitions.partition import Partition


class PartitionDTO(Partition):
    """Represents a Partition Data Transfer Object (DTO) that implements the Partition interface."""

    @unique
    class Type(str, Enum):
        """Type of the partition."""

        RANGE = "range"
        """The range partition type."""
        LIST = "list"
        """The list partition type."""
        IDENTITY = "identity"
        """The identity partition type."""

    @abstractmethod
    def type(self) -> Type:
        """Gets the type of the partition.

        Returns:
            Type: The type of the partition.
        """
        pass  # pragma: no cover



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import config

from gravitino.api.model_change import ModelChange
from gravitino.rest.rest_message import RESTRequest


@dataclass
class ModelUpdateRequestBase(RESTRequest):
    """Base class for all model update requests."""

    _type: str = field(metadata=config(field_name="@type"))

    def __init__(self, action_type: str):
        self._type = action_type

    @abstractmethod
    def model_change(self) -> ModelChange:
        """Convert to model change operation"""
        pass


class ModelUpdateRequest:
    """Namespace for all model update request types."""

    @dataclass
    class UpdateModelNameRequest(ModelUpdateRequestBase):
        """Request to update model name"""

        _new_name: Optional[str] = field(metadata=config(field_name="newName"))

        def __init__(self, new_name: str):
            super().__init__("rename")
            self._new_name = new_name

        def validate(self):
            if not self._new_name:
                raise ValueError('"new_name" field is required')

        def model_change(self) -> ModelChange:
            return ModelChange.rename(self._new_name)

    @dataclass
    class ModelSetPropertyRequest(ModelUpdateRequestBase):
        """Request to set model property"""

        _property: Optional[str] = field(metadata=config(field_name="property"))
        _value: Optional[str] = field(metadata=config(field_name="value"))

        def __init__(self, pro: str, value: str):
            super().__init__("setProperty")
            self._property = pro
            self._value = value

        def validate(self):
            if not self._property:
                raise ValueError('"property" field is required')
            if not self._value:
                raise ValueError('"value" field is required')

        def model_change(self) -> ModelChange:
            return ModelChange.set_property(self._property, self._value)

    @dataclass
    class ModelRemovePropertyRequest(ModelUpdateRequestBase):
        """Request to remove model property"""

        _property: Optional[str] = field(metadata=config(field_name="property"))

        def __init__(self, pro: str):
            super().__init__("removeProperty")
            self._property = pro

        def validate(self):
            if not self._property:
                raise ValueError('"property" field is required')

        def model_change(self) -> ModelChange:
            return ModelChange.remove_property(self._property)

    @dataclass
    class UpdateModelCommentRequest(ModelUpdateRequestBase):
        """Request to update model comment"""

        _new_comment: Optional[str] = field(metadata=config(field_name="newComment"))
        """Represents a request to update the comment on a Metalake."""

        def __init__(self, new_comment: str):
            super().__init__("updateComment")
            self._new_comment = new_comment

        def validate(self):
            """Validates the fields of the request. Always pass."""
            pass

        def model_change(self):
            return ModelChange.update_comment(self._new_comment)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union

from dataclasses_json.core import Json

from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.sorts.sort_order import SortOrder
from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.types.types import Type
from gravitino.dto.rel.partitioning.partitioning import Partitioning
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO

_GravitinoTypeT = TypeVar(
    "_GravitinoTypeT",
    bound=Union[
        Expression, Type, Partitioning, PartitionDTO, Distribution, Index, SortOrder
    ],
)


class JsonSerializable(ABC, Generic[_GravitinoTypeT]):
    """Customized generic Serializer for DataClassJson."""

    @classmethod
    @abstractmethod
    def serialize(cls, data_type: _GravitinoTypeT) -> Json:
        """To serialize the given `data`.

        Args:
            data (_GravitinoTypeT): The data to be serialized.

        Returns:
            Json: The serialized data.
        """
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, data: Json) -> _GravitinoTypeT:
        """To deserialize the given `data`.

        Args:
            data (Json): The data to be deserialized.

        Returns:
            _GravitinoTypeT: The deserialized data.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Optional
from dataclasses import dataclass


@dataclass
class OAuth2ClientCredentialRequest:
    grant_type: str
    client_id: Optional[str]
    client_secret: str
    scope: str

    def to_dict(self, **kwarg):
        return {k: v for k, v in self.__dict__.items() if v is not None}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Dict, List, Optional

from gravitino.api.credential.credential import Credential
from gravitino.api.credential.supports_credentials import SupportsCredentials
from gravitino.api.file.fileset import Fileset
from gravitino.api.metadata_object import MetadataObject
from gravitino.api.metadata_objects import MetadataObjects
from gravitino.client.metadata_object_credential_operations import (
    MetadataObjectCredentialOperations,
)
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.fileset_dto import FilesetDTO
from gravitino.namespace import Namespace
from gravitino.utils import HTTPClient


class GenericFileset(Fileset, SupportsCredentials):
    _fileset: FilesetDTO
    """The fileset data transfer object"""

    _object_credential_operations: MetadataObjectCredentialOperations
    """The metadata object credential operations"""

    def __init__(
        self, fileset: FilesetDTO, rest_client: HTTPClient, full_namespace: Namespace
    ):
        self._fileset = fileset
        metadata_object = MetadataObjects.of(
            [full_namespace.level(1), full_namespace.level(2), fileset.name()],
            MetadataObject.Type.FILESET,
        )
        self._object_credential_operations = MetadataObjectCredentialOperations(
            full_namespace.level(0), metadata_object, rest_client
        )

    def name(self) -> str:
        return self._fileset.name()

    def type(self) -> Fileset.Type:
        return self._fileset.type()

    def storage_locations(self) -> Dict[str, str]:
        return self._fileset.storage_locations()

    def comment(self) -> Optional[str]:
        return self._fileset.comment()

    def properties(self) -> Dict[str, str]:
        return self._fileset.properties()

    def audit_info(self) -> AuditDTO:
        return self._fileset.audit_info()

    def support_credentials(self) -> SupportsCredentials:
        return self

    def get_credentials(self) -> List[Credential]:
        return self._object_credential_operations.get_credentials()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import List

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.named_reference import NamedReference
from gravitino.api.rel.partitions.partition import Partition
from gravitino.api.rel.partitions.partitions import Partitions


class Transform(Expression, ABC):
    """Represents a transform function in the public logical expression API.

    For example, the transform date(ts) is used to derive a date value from a timestamp column.
    The transform name is "date" and its argument is a reference to the "ts" column.
    """

    @abstractmethod
    def name(self) -> str:
        """Gets the transform function name.

        Returns:
            str: The transform function name.
        """
        pass

    @abstractmethod
    def arguments(self) -> List[Expression]:
        """Gets the arguments passed to the transform function.

        Returns:
            List[Expression]: The arguments passed to the transform function.
        """
        pass

    def assignments(self) -> List[Partition]:
        """Gets the preassigned partitions in the partitioning.

        Currently, only `Transforms.ListTransform` and `Transforms.RangeTransform` need to deal with
        assignments

        Returns:
            List[Partition]: The preassigned partitions in the partitioning.
        """
        return Partitions.EMPTY_PARTITIONS

    def children(self) -> List[Expression]:
        return self.arguments()


class SingleFieldTransform(Transform):
    """Base class for transforms on a single field."""

    def __init__(self, ref: NamedReference):
        self.ref = ref

    def field_name(self) -> List[str]:
        """Gets the referenced field name as a list of string parts.

        Returns:
            List[str]: The referenced field name as an array of String parts.
        """
        return self.ref.field_name()

    def references(self) -> List[NamedReference]:
        return [self.ref]

    def arguments(self) -> List[Expression]:
        return [self.ref]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SingleFieldTransform):
            return False
        return self.ref == other.ref

    def __hash__(self) -> int:
        return hash(self.ref)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC


class CatalogChange(ABC):
    """
    A catalog change is a change to a catalog. It can be used to rename a catalog, update the comment
    of a catalog, set a property and value pair for a catalog, or remove a property from a catalog.
    """

    @staticmethod
    def rename(new_name):
        """Creates a new catalog change to rename the catalog.

        Args:
            new_name: The new name of the catalog.

        Returns:
            The catalog change.
        """
        return CatalogChange.RenameCatalog(new_name)

    @staticmethod
    def update_comment(new_comment):
        """Creates a new catalog change to update the catalog comment.

        Args:
            new_comment: The new comment for the catalog.

        Returns:
            The catalog change.
        """
        return CatalogChange.UpdateCatalogComment(new_comment)

    @staticmethod
    def set_property(catalog_property, value):
        """Creates a new catalog change to set the property and value for the catalog.

        Args:
            catalog_property: The property name to set.
            value: The value to set the property to.

        Returns:
            The catalog change.
        """
        return CatalogChange.SetProperty(catalog_property, value)

    @staticmethod
    def remove_property(catalog_property):
        """Creates a new catalog change to remove a property from the catalog.

        Args:
            catalog_property: The property name to remove.

        Returns:
            The catalog change.
        """
        return CatalogChange.RemoveProperty(catalog_property)

    class RenameCatalog:
        """A catalog change to rename the catalog."""

        def __init__(self, new_name):
            self._new_name = new_name

        def new_name(self):
            """Retrieves the new name set for the catalog.

            Returns:
                The new name of the catalog.
            """
            return self._new_name

        def __eq__(self, other) -> bool:
            """Compares this RenameCatalog instance with another object for equality. Two instances are
            considered equal if they designate the same new name for the catalog.

            Args:
                other: The object to compare with this instance.

            Returns:
                true if the given object represents an identical catalog renaming operation; false otherwise.
            """
            if not isinstance(other, CatalogChange.RenameCatalog):
                return False
            return self.new_name() == other.new_name()

        def __hash__(self):
            """Generates a hash code for this RenameCatalog instance. The hash code is primarily based on
            the new name for the catalog.

            Returns:
                A hash code value for this renaming operation.
            """
            return hash(self.new_name())

        def __str__(self):
            """Provides a string representation of the RenameCatalog instance. This string includes the
            class name followed by the new name of the catalog.

            Returns:
                A string summary of this renaming operation.
            """
            return f"RENAMECATALOG {self.new_name()}"

    class UpdateCatalogComment:
        """A catalog change to update the catalog comment."""

        def __init__(self, new_comment):
            self._new_comment = new_comment

        def new_comment(self):
            """Retrieves the new comment intended for the catalog.

            Returns:
                The new comment that has been set for the catalog.
            """
            return self._new_comment

        def __eq__(self, other) -> bool:
            """Compares this UpdateCatalogComment instance with another object for equality.
            Two instances are considered equal if they designate the same new comment for the catalog.

            Args:
                other: The object to compare with this instance.

            Returns:
                true if the given object represents the same comment update; false otherwise.
            """
            if not isinstance(other, CatalogChange.UpdateCatalogComment):
                return False
            return self.new_comment() == other.new_comment()

        def __hash__(self):
            """Generates a hash code for this UpdateCatalogComment instance.
            The hash code is based on the new comment for the catalog.

            Returns:
                A hash code representing this comment update operation.
            """
            return hash(self.new_comment())

        def __str__(self):
            """Provides a string representation of the UpdateCatalogComment instance.
            This string format includes the class name followed by the new comment for the catalog.

            Returns:
                A string summary of this comment update operation.
            """
            return f"UPDATECATALOGCOMMENT {self.new_comment()}"

    class SetProperty:
        """A catalog change to set the property and value for the catalog."""

        def __init__(self, catalog_property, value):
            self._property = catalog_property
            self._value = value

        def property(self):
            """Retrieves the name of the property being set in the catalog.

            Returns:
                The name of the property.
            """
            return self._property

        def value(self):
            """Retrieves the value assigned to the property in the catalog.

            Returns:
                The value of the property.
            """
            return self._value

        def __eq__(self, other) -> bool:
            """Compares this SetProperty instance with another object for equality.
            Two instances are considered equal if they have the same property and value for the catalog.

            Args:
                other: The object to compare with this instance.

            Returns:
                true if the given object represents the same property setting; false otherwise.
            """
            if not isinstance(other, CatalogChange.SetProperty):
                return False
            return self.property() == other.property() and self.value() == other.value()

        def __hash__(self):
            """Generates a hash code for this SetProperty instance.
            The hash code is based on both the property name and its assigned value.

            Returns:
                 A hash code value for this property setting.
            """
            return hash((self.property(), self.value()))

        def __str__(self):
            """Provides a string representation of the SetProperty instance.
            This string format includes the class name followed by the property and its value.

            Returns:
                 A string summary of the property setting.
            """
            return f"SETPROPERTY {self.property()} {self.value()}"

    class RemoveProperty:
        """A catalog change to remove a property from the catalog."""

        def __init__(self, catalog_property):
            self._property = catalog_property

        def get_property(self):
            """Retrieves the name of the property to be removed from the catalog.

            Returns:
                 The name of the property for removal.
            """
            return self._property

        def __eq__(self, other) -> bool:
            """Compares this RemoveProperty instance with another object for equality.
            Two instances are considered equal if they target the same property for removal from the catalog.

            Args:
                other The object to compare with this instance.

            Returns:
                true if the given object represents the same property removal; false otherwise.
            """
            if not isinstance(other, CatalogChange.RemoveProperty):
                return False
            return self._property == other._property

        def __hash__(self):
            """Generates a hash code for this RemoveProperty instance.
            The hash code is based on the property name that is to be removed from the catalog.

            Returns:
                 A hash code value for this property removal operation.
            """
            return hash(self._property)

        def __str__(self):
            """Provides a string representation of the RemoveProperty instance.
            This string format includes the class name followed by the property name to be removed.

            Returns:
                 A string summary of the property removal operation.
            """
            return f"REMOVEPROPERTY {self._property}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import config

from gravitino.rest.rest_message import RESTRequest


@dataclass
class JobRunRequest(RESTRequest):
    """Represents a request to run a job using a specified job template."""

    _job_template_name: str = field(metadata=config(field_name="jobTemplateName"))
    _job_conf: Optional[Dict[str, str]] = field(
        default=None, metadata=config(field_name="jobConf")
    )

    def job_template_name(self) -> str:
        """Returns the job template name to use for running the job."""
        return self._job_template_name

    def job_conf(self) -> Optional[Dict[str, str]]:
        """Returns the job configuration parameters."""
        return self._job_conf

    def validate(self) -> None:
        """Validates the request. Raises ValueError if invalid."""
        if self._job_template_name is None or not self._job_template_name.strip():
            raise ValueError('"jobTemplateName" is required and cannot be empty')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import TYPE_CHECKING, List

from gravitino.utils.precondition import Precondition

if TYPE_CHECKING:
    from gravitino.dto.rel.column_dto import ColumnDTO


class PartitionUtils:
    """Validates the existence of the partition field in the table."""

    @staticmethod
    def validate_field_existence(
        columns: List["ColumnDTO"], field_name: List[str]
    ) -> None:
        """Validates the existence of the partition field in the table.

        Args:
            columns (List[ColumnDTO]): The columns of the table.
            field_name (List[str]): The name of the field to validate.

        Raises:
            IllegalArgumentException:
                If the field does not exist in the table, this exception is thrown.
        """
        Precondition.check_argument(
            columns is not None and len(columns) > 0, "columns cannot be null or empty"
        )
        # TODO: Need to consider the case sensitivity issues. To be optimized.
        partition_column = [
            c for c in columns if c.name().lower() == field_name[0].lower()
        ]

        Precondition.check_argument(
            len(partition_column) == 1, f"Field '{field_name[0]}' not found in table"
        )
        # TODO: should validate nested fieldName after column type support namedStruct



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.dto.rel.partitions.json_serdes.partition_dto_serdes import (
    PartitionDTOSerdes,
)
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO
from gravitino.rest.rest_message import RESTRequest
from gravitino.utils.precondition import Precondition


@dataclass
class AddPartitionsRequest(RESTRequest):
    """Request to add partitions to a table."""

    _partitions: list[PartitionDTO] = field(
        metadata=config(
            field_name="partitions",
            encoder=lambda items: [
                PartitionDTOSerdes.serialize(item) for item in items
            ],
            decoder=lambda values: [
                PartitionDTOSerdes.deserialize(value) for value in values
            ],
        )
    )

    def validate(self):
        Precondition.check_argument(
            self._partitions is not None, "partitions must not be null"
        )
        Precondition.check_argument(
            len(self._partitions) == 1, "Haven't yet implemented multiple partitions"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List

from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.distributions.strategy import Strategy
from gravitino.dto.rel.column_dto import ColumnDTO
from gravitino.dto.rel.expressions.function_arg import FunctionArg
from gravitino.utils.precondition import Precondition


class DistributionDTO(Distribution):
    """Data transfer object representing distribution information.

    Attributes:
        NONE (DistributionDTO):
            A DistributionDTO instance that represents no distribution.
    """

    NONE: "DistributionDTO"

    def __init__(
        self,
        strategy: Strategy,
        number: int,
        args: List[FunctionArg],
    ):
        Precondition.check_argument(
            number >= -1, "bucketNum must be greater than or equal -1"
        )
        Precondition.check_argument(args is not None, "expressions cannot be null")
        self._strategy = strategy if isinstance(strategy, Strategy) else Strategy.HASH
        self._number = number
        self._args = args

    def args(self) -> List[FunctionArg]:
        """Returns the arguments of the function.

        Returns:
            List[FunctionArg]: The arguments of the function.
        """
        return self._args

    def strategy(self) -> Strategy:
        """Returns the strategy of the distribution.

        Returns:
            Strategy: The strategy of the distribution.
        """
        return self._strategy

    def number(self) -> int:
        """Returns the number of buckets.

        Returns:
            int: The number of buckets.
        """
        return self._number

    def expressions(self) -> List[FunctionArg]:
        """Returns the name of the distribution.

        Returns:
            List[FunctionArg]: The name of the distribution.
        """
        return self._args

    def validate(self, columns: List[ColumnDTO]) -> None:
        """Validates the distribution.

        Args:
            columns (List[ColumnDTO]): The columns to be validated.

        Raises:
            IllegalArgumentException: If the distribution is invalid.
        """

        for expression in self._args:
            expression.validate(columns)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DistributionDTO):
            return False
        return self is other or (
            self._number == other.number()
            and self._args == other.args()
            and self._strategy is other.strategy()
        )

    def __hash__(self) -> int:
        result = hash(tuple(self._args))
        result = 31 * result + self._number
        result = 31 * result + hash(self._strategy) if self._strategy else 0
        return result


DistributionDTO.NONE = DistributionDTO(Strategy.NONE, 0, FunctionArg.EMPTY_ARGS)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from gravitino.constants.error import ErrorConstants
from gravitino.dto.responses.error_response import ErrorResponse
from gravitino.exceptions.base import (
    NoSuchSchemaException,
    NoSuchModelException,
    NoSuchModelVersionException,
    NoSuchModelVersionURINameException,
    NotFoundException,
    ModelAlreadyExistsException,
    ModelVersionAliasesAlreadyExistException,
    AlreadyExistsException,
    CatalogNotInUseException,
    MetalakeNotInUseException,
    NotInUseException,
)
from gravitino.exceptions.handlers.rest_error_handler import RestErrorHandler


class ModelErrorHandler(RestErrorHandler):
    def handle(self, error_response: ErrorResponse):
        error_message = error_response.format_error_message()
        code = error_response.code()
        exception_type = error_response.type()

        if code == ErrorConstants.NOT_FOUND_CODE:
            if exception_type == NoSuchSchemaException.__name__:
                raise NoSuchSchemaException(error_message)
            if exception_type == NoSuchModelException.__name__:
                raise NoSuchModelException(error_message)
            if exception_type == NoSuchModelVersionException.__name__:
                raise NoSuchModelVersionException(error_message)
            if exception_type == NoSuchModelVersionURINameException.__name__:
                raise NoSuchModelVersionURINameException(error_message)

            raise NotFoundException(error_message)

        if code == ErrorConstants.ALREADY_EXISTS_CODE:
            if exception_type == ModelAlreadyExistsException.__name__:
                raise ModelAlreadyExistsException(error_message)
            if exception_type == ModelVersionAliasesAlreadyExistException.__name__:
                raise ModelVersionAliasesAlreadyExistException(error_message)

            raise AlreadyExistsException(error_message)

        if code == ErrorConstants.NOT_IN_USE_CODE:
            if exception_type == CatalogNotInUseException.__name__:
                raise CatalogNotInUseException(error_message)
            if exception_type == MetalakeNotInUseException.__name__:
                raise MetalakeNotInUseException(error_message)

            raise NotInUseException(error_message)

        super().handle(error_response)


MODEL_ERROR_HANDLER = ModelErrorHandler()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Optional
from dataclasses import dataclass, field
from dataclasses_json import config

from gravitino.dto.responses.base_response import BaseResponse
from gravitino.auth.auth_constants import AuthConstants
from gravitino.exceptions.base import IllegalArgumentException


@dataclass
class OAuth2TokenResponse(BaseResponse):
    _access_token: str = field(metadata=config(field_name="access_token"))
    _issue_token_type: Optional[str] = field(
        metadata=config(field_name="issued_token_type")
    )
    _token_type: str = field(metadata=config(field_name="token_type"))
    _expires_in: int = field(metadata=config(field_name="expires_in"))
    _scope: str = field(metadata=config(field_name="scope"))
    _refresh_token: Optional[str] = field(metadata=config(field_name="refresh_token"))

    def validate(self):
        """Validates the response.

        Raise:
            IllegalArgumentException If the response is invalid, this exception is thrown.
        """
        if self._access_token is None:
            raise IllegalArgumentException("Invalid access token: None")

        if (
            AuthConstants.AUTHORIZATION_BEARER_HEADER.strip().lower()
            != self._token_type.lower()
        ):
            raise IllegalArgumentException(
                f'Unsupported token type: {self._token_type} (must be "bearer")'
            )

    def access_token(self) -> str:
        return self._access_token



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Optional, Dict

from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.rest.rest_message import RESTRequest


@dataclass
class MetalakeCreateRequest(RESTRequest):
    """Represents a request to create a Metalake."""

    _name: str = field(metadata=config(field_name="name"))
    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    _properties: Optional[Dict[str, str]] = field(
        metadata=config(field_name="properties")
    )

    def __init__(
        self, name: str = None, comment: str = None, properties: Dict[str, str] = None
    ):
        super().__init__()

        self._name = name.strip() if name else None
        self._comment = comment.strip() if comment else None
        self._properties = properties

    def name(self) -> str:
        return self._name

    def validate(self):
        if not self._name:
            raise ValueError('"name" field is required and cannot be empty')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from enum import Enum

from gravitino.api.rel.expressions.sorts.null_ordering import NullOrdering


class SortDirection(Enum):
    """A sort direction used in sorting expressions.
    Each direction has a default null ordering that is implied if no null ordering is specified explicitly.
    """

    ASCENDING = ("asc", NullOrdering.NULLS_FIRST)
    """Ascending sort direction. Nulls appear first. For ascending order, this means nulls appear at the beginning."""

    DESCENDING = ("desc", NullOrdering.NULLS_LAST)
    """Descending sort direction. Nulls appear last. For ascending order, this means nulls appear at the end."""

    def __init__(self, direction: str, default_null_ordering: NullOrdering):
        self._direction = direction
        self._default_null_ordering = default_null_ordering

    def direction(self) -> str:
        return self._direction

    def default_null_ordering(self) -> NullOrdering:
        """
        Returns the default null ordering to use if no null ordering is specified explicitly.

        Returns:
            NullOrdering: The default null ordering.
        """
        return self._default_null_ordering

    def __str__(self) -> str:
        if self == SortDirection.ASCENDING:
            return SortDirection.ASCENDING.direction()
        if self == SortDirection.DESCENDING:
            return SortDirection.DESCENDING.direction()

        raise ValueError(f"Unexpected sort direction: {self}")

    @staticmethod
    def from_string(direction: str):
        """
        Returns the SortDirection from the string representation.

        Args:
            direction: The string representation of the sort direction.

        Returns:
            SortDirection: The corresponding SortDirection.
        """
        direction = direction.lower()
        if direction == SortDirection.ASCENDING.direction():
            return SortDirection.ASCENDING
        if direction == SortDirection.DESCENDING.direction():
            return SortDirection.DESCENDING

        raise ValueError(f"Unexpected sort direction: {direction}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Dict, List

from gravitino.api.rel.partitions.identity_partition import IdentityPartition
from gravitino.dto.rel.expressions.literal_dto import LiteralDTO
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO


class IdentityPartitionDTO(PartitionDTO, IdentityPartition):
    """Represents an Identity Partition Data Transfer Object (DTO) that implements the IdentityPartition interface."""

    def __init__(
        self,
        name: str,
        values: List[LiteralDTO],
        field_names: List[List[str]],
        properties: Dict[str, str],
    ):
        self._name = name
        self._values = values
        self._field_names = field_names
        self._properties = properties

    def name(self) -> str:
        return self._name

    def values(self) -> List[LiteralDTO]:
        return self._values

    def field_names(self) -> List[List[str]]:
        return self._field_names

    def properties(self) -> Dict[str, str]:
        return self._properties

    def type(self) -> PartitionDTO.Type:
        return self.Type.IDENTITY

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, IdentityPartitionDTO):
            return False
        return (
            self is value
            or self._name == value.name()
            and self._values == value.values()
            and self._field_names == value.field_names()
            and self._properties == value.properties()
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._name,
                tuple(self._values),
                tuple(tuple(field) for field in self._field_names),
                tuple(self._properties.items()),
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.api.file.fileset_change import FilesetChange
from gravitino.rest.rest_message import RESTRequest


@dataclass
class FilesetUpdateRequestBase(RESTRequest):
    _type: str = field(metadata=config(field_name="@type"))

    def __init__(self, action_type: str):
        self._type = action_type

    @abstractmethod
    def fileset_change(self):
        pass


class FilesetUpdateRequest:
    """Request to update a fileset."""

    @dataclass
    class RenameFilesetRequest(FilesetUpdateRequestBase):
        """The fileset update request for renaming a fileset."""

        _new_name: str = field(metadata=config(field_name="newName"))
        """The new name for the Fileset."""

        def __init__(self, new_name: str):
            super().__init__("rename")
            self._new_name = new_name

        def validate(self):
            """Validates the fields of the request.

            Raises:
                IllegalArgumentException if the new name is not set.
            """
            if not self._new_name:
                raise ValueError('"new_name" field is required and cannot be empty')

        def fileset_change(self):
            """Returns the fileset change.

            Returns:
                the fileset change.
            """
            return FilesetChange.rename(self._new_name)

    @dataclass
    class UpdateFilesetCommentRequest(FilesetUpdateRequestBase):
        """Represents a request to update the comment on a Fileset."""

        _new_comment: str = field(metadata=config(field_name="newComment"))
        """The new comment for the Fileset."""

        def __init__(self, new_comment: str):
            super().__init__("updateComment")
            self._new_comment = new_comment

        def validate(self):
            """Validates the fields of the request. Always pass."""
            pass

        def fileset_change(self):
            """Returns the fileset change"""
            return FilesetChange.update_comment(self._new_comment)

    @dataclass
    class SetFilesetPropertyRequest(FilesetUpdateRequestBase):
        """Represents a request to set a property on a Fileset."""

        _property: str = field(metadata=config(field_name="property"))
        """The property to set."""

        _value: str = field(metadata=config(field_name="value"))
        """The value of the property."""

        def __init__(self, fileset_property: str, value: str):
            super().__init__("setProperty")
            self._property = fileset_property
            self._value = value

        def validate(self):
            """Validates the fields of the request.

            Raises:
                 IllegalArgumentException if fileset_property or value are not set.
            """
            if not self._property:
                raise ValueError(
                    '"fileset_property" field is required and cannot be empty'
                )
            if not self._value:
                raise ValueError('"value" field is required and cannot be empty')

        def fileset_change(self):
            return FilesetChange.set_property(self._property, self._value)

    @dataclass
    class RemoveFilesetPropertyRequest(FilesetUpdateRequestBase):
        """Represents a request to remove a property from a Fileset."""

        _property: str = field(metadata=config(field_name="property"))
        """The property to remove."""

        def __init__(self, fileset_property: str):
            super().__init__("removeProperty")
            self._property = fileset_property

        def validate(self):
            """Validates the fields of the request.

            Raises:
                 IllegalArgumentException if fileset_property is not set.
            """
            if not self._property:
                raise ValueError(
                    '"fileset_property" field is required and cannot be empty'
                )

        def fileset_change(self):
            return FilesetChange.remove_property(self._property)

    @dataclass
    class RemoveFilesetCommentRequest(FilesetUpdateRequestBase):
        """Represents a request to remove comment from a Fileset.

        Deprecated:
            Please use `UpdateFilesetCommentRequest` with null value as the argument instead.
        """

        def __init__(self):
            super().__init__("removeComment")

        def validate(self):
            """Validates the fields of the request.

            always pass
            """
            pass

        def fileset_change(self):
            return FilesetChange.remove_comment()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class GVFSConfig:
    CACHE_SIZE = "cache_size"
    DEFAULT_CACHE_SIZE = 20

    CACHE_EXPIRED_TIME = "cache_expired_time"
    DEFAULT_CACHE_EXPIRED_TIME = 3600

    AUTH_TYPE = "auth_type"
    SIMPLE_AUTH_TYPE = "simple"

    OAUTH2_AUTH_TYPE = "oauth2"
    OAUTH2_SERVER_URI = "oauth2_server_uri"
    OAUTH2_CREDENTIAL = "oauth2_credential"
    OAUTH2_PATH = "oauth2_path"
    OAUTH2_SCOPE = "oauth2_scope"

    GVFS_FILESYSTEM_GCS_SERVICE_KEY_FILE = "gcs_service_account_file"

    GVFS_FILESYSTEM_S3_ACCESS_KEY = "s3_access_key_id"
    GVFS_FILESYSTEM_S3_SECRET_KEY = "s3_secret_access_key"
    GVFS_FILESYSTEM_S3_ENDPOINT = "s3_endpoint"

    GVFS_FILESYSTEM_OSS_ACCESS_KEY = "oss_access_key_id"
    GVFS_FILESYSTEM_OSS_SECRET_KEY = "oss_secret_access_key"
    GVFS_FILESYSTEM_OSS_ENDPOINT = "oss_endpoint"

    GVFS_FILESYSTEM_AZURE_ACCOUNT_NAME = "azure_storage_account_name"
    GVFS_FILESYSTEM_AZURE_ACCOUNT_KEY = "azure_storage_account_key"

    # This configuration marks the expired time of the credential. For instance, if the credential
    # fetched from Gravitino server has expired time of 3600 seconds, and the credential_expired_time_ration is 0.5
    # then the credential will be considered as expired after 1800 seconds and will try to retrieve a new credential.
    GVFS_FILESYSTEM_CREDENTIAL_EXPIRED_TIME_RATIO = "credential_expiration_ratio"

    # The default value of the credential_expired_time_ratio is 0.5
    DEFAULT_CREDENTIAL_EXPIRED_TIME_RATIO = 0.5

    # The configuration key for the fileset with multiple locations, on which the file system will operate.
    # The default value is "default".
    GVFS_FILESYSTEM_CURRENT_LOCATION_NAME = "current_location_name"

    # The configuration key for the env variable name that indicates the current location name. If
    # not set, the file system will read the location name from CURRENT_LOCATION_NAME env variable.
    GVFS_FILESYSTEM_CURRENT_LOCATION_NAME_ENV_VAR = "current_location_name_env_var"

    # The configuration key for the class name of the file system operations.
    # The default value is "org.apache.gravitino.filesystem.hadoop.DefaultGVFSOperations".
    GVFS_FILESYSTEM_OPERATIONS = "operations_class"

    # The hook class that will be used to intercept file system operations.
    GVFS_FILESYSTEM_HOOK = "hook_class"

    # The configuration key prefix for the client request headers.
    GVFS_FILESYSTEM_CLIENT_REQUEST_HEADER_PREFIX = "client_request_header_"

    # The configuration key for whether to enable credential vending. The default is false.
    GVFS_FILESYSTEM_ENABLE_CREDENTIAL_VENDING = "enable_credential_vending"

    # The configuration key prefix for the client.
    GVFS_FILESYSTEM_CLIENT_CONFIG_PREFIX = "gvfs_gravitino_client_"

    # The configuration key for whether to enable fileset catalog cache. The default is false.
    # Note that this cache causes a side effect: if you modify the fileset or fileset catalog metadata,
    # the client can not see the latest changes.
    GVFS_FILESYSTEM_ENABLE_FILESET_METADATA_CACHE = "enable_fileset_metadata_cache"

    # The configuration key for whether to enable auto-creation of fileset location when the
    # server-side filesystem ops are disabled and the location does not exist. The default is true.
    GVFS_FILESYSTEM_AUTO_CREATE_LOCATION = "auto_create_location"

    # The configuration prefix for user-defined path-specific configurations.
    # Note: location names must NOT contain ('.', '_'),
    # because '_' is used as the separator between prefix, location name, and property name.
    # Configuration is defined in two steps:
    #   1) Map a logical location name to a base location:
    #        fs_path_config_<location_name>=<base_location>
    #   2) Define properties for that logical location:
    #        fs_path_config_<location_name>_<property_name>=<property_value>
    # Example:
    #   fs_path_config_my-s3-location=s3://bucket/path
    #   fs_path_config_my-s3-location_aws-access-key=XXX
    FS_GRAVITINO_PATH_CONFIG_PREFIX = "fs_path_config_"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import ABC
from typing import List, Dict, Optional, Any


class JobTemplateChange:
    """
    The interface for job template changes. A job template change is an operation that modifies a job
    template. It can be one of the following:

      - Rename the job template.
      - Update the comment of the job template.
      - Update the job template details, such as executable, arguments, environments, custom
        fields, etc.
    """

    @staticmethod
    def rename(new_name: str) -> "JobTemplateChange":
        """
        Creates a new job template change to update the name of the job template.

        Args:
            new_name: The new name of the job template.

        Returns:
            The job template change.
        """
        return RenameJobTemplate(new_name)

    @staticmethod
    def update_comment(new_comment: str) -> "JobTemplateChange":
        """
        Creates a new job template change to update the comment of the job template.

        Args:
            new_comment: The new comment of the job template.

        Returns:
            The job template change.
        """
        return UpdateJobTemplateComment(new_comment)

    @staticmethod
    def update_template(template_update: "TemplateUpdate") -> "JobTemplateChange":
        """
        Creates a new job template change to update the details of the job template.

        Args:
            template_update: The template update details.

        Returns:
            The job template change.
        """
        return UpdateJobTemplate(template_update)


class RenameJobTemplate(JobTemplateChange):
    """
    A job template change to rename the job template.
    """

    def __init__(self, new_name: str):
        """
        Initialize a RenameJobTemplate change.

        Args:
            new_name: The new name of the job template.
        """
        self._new_name = new_name

    def get_new_name(self) -> str:
        """
        Get the new name of the job template.

        Returns:
            The new name of the job template.
        """
        return self._new_name

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, RenameJobTemplate) and self._new_name == other._new_name
        )

    def __hash__(self) -> int:
        return hash(self._new_name)

    def __str__(self) -> str:
        return f"RENAME JOB TEMPLATE {self._new_name}"


class UpdateJobTemplateComment(JobTemplateChange):
    """
    A job template change to update the comment of the job template.
    """

    def __init__(self, new_comment: str):
        """
        Initialize an UpdateJobTemplateComment change.

        Args:
            new_comment: The new comment of the job template.
        """
        self._new_comment = new_comment

    def get_new_comment(self) -> str:
        """
        Get the new comment of the job template.

        Returns:
            The new comment of the job template.
        """
        return self._new_comment

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, UpdateJobTemplateComment)
            and self._new_comment == other._new_comment
        )

    def __hash__(self) -> int:
        return hash(self._new_comment)

    def __str__(self) -> str:
        return f"UPDATE JOB TEMPLATE COMMENT {self._new_comment}"


class UpdateJobTemplate(JobTemplateChange):
    """
    A job template change to update the details of the job template.
    """

    def __init__(self, template_update: "TemplateUpdate"):
        """
        Initialize an UpdateJobTemplate change.

        Args:
            template_update: The job template update details.
        """
        self._template_update = template_update

    def get_template_update(self) -> "TemplateUpdate":
        """
        Get the job template update.

        Returns:
            The job template update.
        """
        return self._template_update

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, UpdateJobTemplate)
            and self._template_update == other._template_update
        )

    def __hash__(self) -> int:
        return hash(self._template_update)

    def __str__(self) -> str:
        return f"UPDATE JOB TEMPLATE {type(self._template_update).__name__}"


class TemplateUpdate(ABC):
    """
    Base class for template updates.
    """

    def __init__(
        self,
        new_executable: Optional[str] = None,
        new_arguments: Optional[List[str]] = None,
        new_environments: Optional[Dict[str, str]] = None,
        new_custom_fields: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize a TemplateUpdate.

        Args:
            new_executable: The new executable of the job template.
            new_arguments: The new arguments of the job template.
            new_environments: The new environments of the job template.
            new_custom_fields: The new custom fields of the job template.
        """
        self._new_executable = new_executable
        self._new_arguments = new_arguments
        self._new_environments = new_environments
        self._new_custom_fields = new_custom_fields

    def get_new_executable(self) -> Optional[str]:
        """
        Get the new executable of the job template.

        Returns:
            The new executable of the job template.
        """
        return self._new_executable

    def get_new_arguments(self) -> Optional[List[str]]:
        """
        Get the new arguments of the job template.

        Returns:
            The new arguments of the job template.
        """
        return self._new_arguments

    def get_new_environments(self) -> Optional[Dict[str, str]]:
        """
        Get the new environments of the job template.

        Returns:
            The new environments of the job template.
        """
        return self._new_environments

    def get_new_custom_fields(self) -> Optional[Dict[str, str]]:
        """
        Get the new custom fields of the job template.

        Returns:
            The new custom fields of the job template.
        """
        return self._new_custom_fields

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TemplateUpdate):
            return False
        return (
            self._new_executable == other._new_executable
            and self._new_arguments == other._new_arguments
            and self._new_environments == other._new_environments
            and self._new_custom_fields == other._new_custom_fields
        )

    def __hash__(self) -> int:
        return hash(
            (
                self._new_executable,
                tuple(self._new_arguments),
                frozenset(self._new_environments.items()),
                frozenset(self._new_custom_fields.items()),
            )
        )


class ShellTemplateUpdate(TemplateUpdate):
    """
    A template update for shell job templates.
    """

    def __init__(
        self,
        new_executable: Optional[str] = None,
        new_arguments: Optional[List[str]] = None,
        new_environments: Optional[Dict[str, str]] = None,
        new_custom_fields: Optional[Dict[str, str]] = None,
        new_scripts: Optional[List[str]] = None,
    ):
        """
        Initialize a ShellTemplateUpdate.

        Args:
            new_executable: The new executable of the shell job template.
            new_arguments: The new arguments of the shell job template.
            new_environments: The new environments of the shell job template.
            new_custom_fields: The new custom fields of the shell job template.
            new_scripts: The new scripts of the shell job template.
        """
        super().__init__(
            new_executable, new_arguments, new_environments, new_custom_fields
        )
        self._new_scripts = new_scripts

    def get_new_scripts(self) -> Optional[List[str]]:
        """
        Get the new scripts of the shell job template.

        Returns:
            The new scripts of the shell job template.
        """
        return self._new_scripts

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ShellTemplateUpdate)
            and super().__eq__(other)
            and self._new_scripts == other._new_scripts
        )

    def __hash__(self) -> int:
        return hash((super().__hash__(), tuple(self._new_scripts)))


class SparkTemplateUpdate(TemplateUpdate):
    """
    A template update for spark job templates.
    """

    def __init__(
        self,
        new_executable: Optional[str] = None,
        new_arguments: Optional[List[str]] = None,
        new_environments: Optional[Dict[str, str]] = None,
        new_custom_fields: Optional[Dict[str, str]] = None,
        new_class_name: Optional[str] = None,
        new_jars: Optional[List[str]] = None,
        new_files: Optional[List[str]] = None,
        new_archives: Optional[List[str]] = None,
        new_configs: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize a SparkTemplateUpdate.

        Args:
            new_executable: The new executable of the spark job template.
            new_arguments: The new arguments of the spark job template.
            new_environments: The new environments of the spark job template.
            new_custom_fields: The new custom fields of the spark job template.
            new_class_name: The new class name of the spark job template.
            new_jars: The new jars of the spark job template.
            new_files: The new files of the spark job template.
            new_archives: The new archives of the spark job template.
            new_configs: The new configs of the spark job template.
        """
        super().__init__(
            new_executable, new_arguments, new_environments, new_custom_fields
        )
        self._new_class_name = new_class_name
        self._new_jars = new_jars
        self._new_files = new_files
        self._new_archives = new_archives
        self._new_configs = new_configs

    def get_new_class_name(self) -> Optional[str]:
        """
        Get the new class name of the spark job template.

        Returns:
            The new class name of the spark job template.
        """
        return self._new_class_name

    def get_new_jars(self) -> Optional[List[str]]:
        """
        Get the new jars of the spark job template.

        Returns:
            The new jars of the spark job template.
        """
        return self._new_jars

    def get_new_files(self) -> Optional[List[str]]:
        """
        Get the new files of the spark job template.

        Returns:
            The new files of the spark job template.
        """
        return self._new_files

    def get_new_archives(self) -> Optional[List[str]]:
        """
        Get the new archives of the spark job template.

        Returns:
            The new archives of the spark job template.
        """
        return self._new_archives

    def get_new_configs(self) -> Optional[Dict[str, str]]:
        """
        Get the new configs of the spark job template.

        Returns:
            The new configs of the spark job template.
        """
        return self._new_configs

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, SparkTemplateUpdate)
            and super().__eq__(other)
            and self._new_class_name == other._new_class_name
            and self._new_jars == other._new_jars
            and self._new_files == other._new_files
            and self._new_archives == other._new_archives
            and self._new_configs == other._new_configs
        )

    def __hash__(self) -> int:
        return hash(
            (
                super().__hash__(),
                self._new_class_name,
                tuple(self._new_jars),
                tuple(self._new_files),
                tuple(self._new_archives),
                frozenset(self._new_configs.items()),
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import List
from gravitino.api.credential.credential import Credential
from gravitino.exceptions.base import (
    NoSuchCredentialException,
    IllegalStateException,
)


class SupportsCredentials(ABC):
    """Represents interface to get credentials."""

    @abstractmethod
    def get_credentials(self) -> List[Credential]:
        """Retrieves a List of Credential objects.

        Returns:
            A List of Credential objects. In most cases the array only contains
        one credential. If the object like Fileset contains multiple locations
        for different storages like HDFS, S3, the array will contain multiple
        credentials. The array could be empty if you request a credential for
        a catalog but the credential provider couldn't generate the credential
        for the catalog, like S3 token credential provider only generate
        credential for the specific object like Fileset,Table. There will be at
        most one credential for one credential type.
        """
        pass

    def get_credential(self, credential_type: str) -> Credential:
        """Retrieves Credential object based on the specified credential type.

        Args:
            credential_type: The type of the credential like s3-token,
            s3-secret-key which are defined in the specific credentials.
        Returns:
            An Credential object with the specified credential type.
        Raises:
            NoSuchCredentialException If the specific credential cannot be found.
            IllegalStateException if multiple credential can be found.
        """

        credentials = self.get_credentials()
        matched_credentials = [
            credential
            for credential in credentials
            if credential.credential_type == credential_type
        ]
        if len(matched_credentials) == 0:
            raise NoSuchCredentialException(
                f"No credential found for the credential type: {credential_type}"
            )
        if len(matched_credentials) > 1:
            raise IllegalStateException(
                f"Multiple credentials found for the credential type: {credential_type}"
            )
        return matched_credentials[0]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import List

from dataclasses_json import config

from gravitino.dto.metalake_dto import MetalakeDTO
from gravitino.dto.responses.base_response import BaseResponse


@dataclass
class MetalakeListResponse(BaseResponse):
    """Represents a response containing a list of metalakes."""

    _metalakes: List[MetalakeDTO] = field(metadata=config(field_name="metalakes"))

    def metalakes(self) -> List[MetalakeDTO]:
        return self._metalakes

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException if catalog identifiers are not set.
        """
        super().validate()

        if self._metalakes is None:
            raise ValueError("metalakes must be non-null")

        for metalake in self._metalakes:
            if not metalake.name():
                raise ValueError("metalake 'name' must not be null and empty")
            if not metalake.audit_info():
                raise ValueError("metalake 'audit' must not be null")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import importlib
import logging
import sys
import time
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Tuple, List, Dict, Any

from fsspec import AbstractFileSystem
from fsspec.implementations.arrow import ArrowFSWrapper
from fsspec.implementations.local import LocalFileSystem
from fsspec.utils import infer_storage_options

from gravitino.api.credential.adls_token_credential import ADLSTokenCredential
from gravitino.api.credential.azure_account_key_credential import (
    AzureAccountKeyCredential,
)
from gravitino.api.credential.credential import Credential
from gravitino.api.credential.gcs_token_credential import GCSTokenCredential
from gravitino.api.credential.oss_secret_key_credential import OSSSecretKeyCredential
from gravitino.api.credential.oss_token_credential import OSSTokenCredential
from gravitino.api.credential.s3_secret_key_credential import S3SecretKeyCredential
from gravitino.api.credential.s3_token_credential import S3TokenCredential
from gravitino.exceptions.base import GravitinoRuntimeException
from gravitino.filesystem.gvfs_config import GVFSConfig

TIME_WITHOUT_EXPIRATION = sys.maxsize
SLASH = "/"

logger = logging.getLogger(__name__)


class StorageType(Enum):
    HDFS = "hdfs"
    LOCAL = "file"
    GCS = "gs"
    S3A = "s3a"
    OSS = "oss"
    ABS = "abfss"


class StorageHandler(ABC):
    """A handler class that holds the information about the actual file location and the file system which used in
    the GravitinoVirtualFileSystem's operations.
    """

    @abstractmethod
    def storage_type(self) -> StorageType:
        """Get the storage type.
        :return: The storage type
        """
        pass

    @abstractmethod
    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        """Get the file system.
        :return: The file system
        """
        pass

    @abstractmethod
    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        """Get the file system with expiration.
        :return: The file system and the expiration time
        """
        pass

    def actual_info_to_gvfs_info(
        self,
        entry: Dict,
        fileset_location: str,
        gvfs_path_prefix: str,
    ):
        """Convert a file info from an actual entry to a virtual entry.
        :param entry: A dict of the actual file info
        :param fileset_location: Storage location of the fileset
        :param gvfs_path_prefix: The prefix of the virtual path
        :return A dict of the virtual file info
        """
        path = self.actual_path_to_gvfs_path(
            entry["name"], fileset_location, gvfs_path_prefix
        )
        last_modified = self._get_last_modified(entry)
        if last_modified is None:
            logger.info(
                "The last modified time is %s for %s", last_modified, fileset_location
            )
        return {
            "name": path,
            "size": entry["size"],
            "type": entry["type"],
            "mtime": last_modified,
        }

    def actual_path_to_gvfs_path(
        self,
        actual_path: str,
        fileset_location: str,
        gvfs_path_prefix: str,
    ):
        """Convert the actual path to the gvfs path.
        :param actual_path: The actual path
        :param fileset_location: The fileset location
        :param gvfs_path_prefix: The gvfs path prefix
        :return: The gvfs path
        """
        if not fileset_location.startswith(f"{self.storage_type().value}:/"):
            raise GravitinoRuntimeException(
                f"Fileset location {fileset_location} is not a {self.storage_type()} path."
            )
        actual_prefix = self._get_actual_prefix(fileset_location)
        normalized_actual_path = self._normalize_actual_path_if_needed(
            infer_storage_options(actual_prefix), actual_path
        )
        if not actual_path.startswith(actual_prefix):
            raise GravitinoRuntimeException(
                f"Path {actual_path} does not start with valid prefix {actual_prefix}."
            )

        # if the fileset location is end with "/",
        # we should truncate this to avoid replace issues.
        if actual_prefix.endswith(SLASH) and not gvfs_path_prefix.endswith(SLASH):
            return f"{normalized_actual_path.replace(actual_prefix[:-1], gvfs_path_prefix)}"
        return f"{normalized_actual_path.replace(actual_prefix, gvfs_path_prefix)}"

    @abstractmethod
    def _get_actual_prefix(self, actual_path: str) -> str:
        """Get the actual path prefix from the actual path.
        :param actual_path: The actual path
        :return: The actual path prefix
        """
        pass

    @abstractmethod
    def _get_most_suitable_credential(self, credentials: List[Credential]):
        """Get the most suitable credential.
        :param credentials: The credentials
        :return: The most suitable credential
        """
        pass

    def strip_storage_protocol(self, path: str) -> str:
        """Strip the storage protocol from the path.
          Before passing the path to the underlying file system for processing,
           pre-process the protocol information in the path.
          Some file systems require special processing.
          For HDFS, we can pass the path like 'hdfs://{host}:{port}/xxx'.
          For Local, we can pass the path like '/tmp/xxx'.
        :param path: The path
        :return: The stripped path
        """
        return path

    def _get_expire_time_by_ratio(self, expire_time: int, options: Dict[str, str]):
        if expire_time <= 0:
            return TIME_WITHOUT_EXPIRATION

        ratio = float(
            options.get(
                GVFSConfig.GVFS_FILESYSTEM_CREDENTIAL_EXPIRED_TIME_RATIO,
                GVFSConfig.DEFAULT_CREDENTIAL_EXPIRED_TIME_RATIO,
            )
            if options
            else GVFSConfig.DEFAULT_CREDENTIAL_EXPIRED_TIME_RATIO
        )
        return time.time() * 1000 + (expire_time - time.time() * 1000) * ratio

    def _normalize_actual_path_if_needed(
        self, storage_options: Dict[str, Any], actual_path: str
    ) -> str:
        return actual_path

    def _get_last_modified(self, entry: Dict):
        return entry["mtime"] if "mtime" in entry else None


class LocalStorageHandler(StorageHandler):
    """Local Storage Handler"""

    def storage_type(self):
        return StorageType.LOCAL

    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        return LocalFileSystem(**kwargs)

    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        """Get the file system with expiration.
        :return: The file system and the expiration time
        """
        return TIME_WITHOUT_EXPIRATION, self.get_filesystem(actual_path, **kwargs)

    def _get_actual_prefix(
        self,
        actual_path: str,
    ) -> str:
        return actual_path[len(f"{self.storage_type().value}:") :]

    def _get_most_suitable_credential(self, credentials: List[Credential]):
        return None

    def strip_storage_protocol(self, path: str):
        """Strip the storage protocol from the path.
        :param path: The path
        :return: The stripped path
        """
        return path[len(f"{self.storage_type().value}:") :]


class HDFSStorageHandler(StorageHandler):
    """HDFS Storage Handler"""

    def storage_type(self):
        return StorageType.HDFS

    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        fs_class = importlib.import_module("pyarrow.fs").HadoopFileSystem
        return ArrowFSWrapper(fs_class.from_uri(actual_path))

    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        """Get the file system with expiration.
        :return: The file system and the expiration time
        """
        return TIME_WITHOUT_EXPIRATION, self.get_filesystem(actual_path, **kwargs)

    def _get_actual_prefix(
        self,
        actual_path: str,
    ) -> str:
        return infer_storage_options(actual_path)["path"]

    def _get_most_suitable_credential(self, credentials: List[Credential]):
        return None


class S3StorageHandler(StorageHandler):
    """S3 Storage Handler"""

    def storage_type(self):
        return StorageType.S3A

    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        """Get the file system with expiration.
        :return: The file system and the expiration time
        """
        # S3 endpoint from client options has a higher priority, override the endpoint from catalog properties.
        # Note: the endpoint may not be a real S3 endpoint, it can be a simulated S3 endpoint, such as minio,
        # so though the endpoint is not a required field for S3FileSystem, we still need to assign the endpoint
        # to the S3FileSystem
        s3_endpoint = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_S3_ENDPOINT
        ) or catalog_props.get("s3-endpoint")

        if credentials:
            credential = self._get_most_suitable_credential(credentials)
            if credential is not None:
                expire_time = self._get_expire_time_by_ratio(
                    credential.expire_time_in_ms(),
                    catalog_props,
                )
                if isinstance(credential, S3TokenCredential):
                    fs = self.get_filesystem(
                        key=credential.access_key_id(),
                        secret=credential.secret_access_key(),
                        token=credential.session_token(),
                        endpoint_url=s3_endpoint,
                        **kwargs,
                    )
                    return expire_time, fs
                if isinstance(credential, S3SecretKeyCredential):
                    fs = self.get_filesystem(
                        key=credential.access_key_id(),
                        secret=credential.secret_access_key(),
                        endpoint_url=s3_endpoint,
                        **kwargs,
                    )
                    return expire_time, fs

        # this is the old way to get the s3 file system
        # get 'aws_access_key_id' from s3_options, if the key is not found, throw an exception
        aws_access_key_id = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_S3_ACCESS_KEY, None
        )
        if aws_access_key_id is None:
            raise GravitinoRuntimeException(
                "AWS access key id is not found in the options."
            )

        # get 'aws_secret_access_key' from s3_options, if the key is not found, throw an exception
        aws_secret_access_key = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_S3_SECRET_KEY
        )
        if aws_secret_access_key is None:
            raise GravitinoRuntimeException(
                "AWS secret access key is not found in the options."
            )

        return (
            TIME_WITHOUT_EXPIRATION,
            self.get_filesystem(
                key=aws_access_key_id,
                secret=aws_secret_access_key,
                endpoint_url=s3_endpoint,
                **kwargs,
            ),
        )

    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        return importlib.import_module("s3fs").S3FileSystem(**kwargs)

    def _get_actual_prefix(
        self,
        actual_path: str,
    ) -> str:
        return infer_storage_options(actual_path)["path"]

    def _get_most_suitable_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, use the
            # secret key credential.
            if isinstance(credential, S3TokenCredential):
                return credential

        for credential in credentials:
            if isinstance(credential, S3SecretKeyCredential):
                return credential
        return None

    def _get_last_modified(self, entry: Dict):
        return entry["LastModified"] if "LastModified" in entry else None


class GCSStorageHandler(StorageHandler):
    """GCS Storage Handler"""

    def storage_type(self):
        return StorageType.GCS

    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        return importlib.import_module("gcsfs").GCSFileSystem(**kwargs)

    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        if credentials:
            credential = self._get_most_suitable_credential(credentials)
            if credential is not None:
                expire_time = self._get_expire_time_by_ratio(
                    credential.expire_time_in_ms(),
                    catalog_props,
                )
                if isinstance(credential, GCSTokenCredential):
                    return (
                        expire_time,
                        self.get_filesystem(
                            token=credential.token(),
                            **kwargs,
                        ),
                    )

        # get 'service-account-key' from gcs_options, if the key is not found, throw an exception
        service_account_key_path = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_GCS_SERVICE_KEY_FILE
        )
        if service_account_key_path is None:
            raise GravitinoRuntimeException(
                "Service account key is not found in the options."
            )
        return (
            TIME_WITHOUT_EXPIRATION,
            self.get_filesystem(
                token=service_account_key_path,
                **kwargs,
            ),
        )

    def _get_actual_prefix(
        self,
        actual_path: str,
    ) -> str:
        return infer_storage_options(actual_path)["path"]

    def _get_most_suitable_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, return None.
            if isinstance(credential, GCSTokenCredential):
                return credential
        return None


class OSSStorageHandler(StorageHandler):
    """OSS Storage Handler"""

    def storage_type(self):
        return StorageType.OSS

    def strip_storage_protocol(self, path: str):
        """Strip the storage protocol from the path.
        if we do not remove the protocol, it will always return an empty array.
        :param path: The path
        :return: The stripped path
        """
        if path.startswith(f"{self.storage_type().value}://"):
            return path[len(f"{self.storage_type().value}://") :]
        return path

    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        return importlib.import_module("ossfs").OSSFileSystem(**kwargs)

    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        # OSS endpoint from client options has a higher priority, override the endpoint from catalog properties.
        oss_endpoint = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_OSS_ENDPOINT
        ) or catalog_props.get("oss-endpoint")

        if credentials:
            credential = self._get_most_suitable_credential(credentials)
            if credential is not None:
                expire_time = self._get_expire_time_by_ratio(
                    credential.expire_time_in_ms(),
                    catalog_props,
                )
                if isinstance(credential, OSSTokenCredential):
                    fs = self.get_filesystem(
                        key=credential.access_key_id(),
                        secret=credential.secret_access_key(),
                        token=credential.security_token(),
                        endpoint=oss_endpoint,
                        **kwargs,
                    )
                    return expire_time, fs
                if isinstance(credential, OSSSecretKeyCredential):
                    return (
                        expire_time,
                        self.get_filesystem(
                            key=credential.access_key_id(),
                            secret=credential.secret_access_key(),
                            endpoint=oss_endpoint,
                            **kwargs,
                        ),
                    )

        # get 'oss_access_key_id' from oss options, if the key is not found, throw an exception
        oss_access_key_id = catalog_props.get(GVFSConfig.GVFS_FILESYSTEM_OSS_ACCESS_KEY)
        if oss_access_key_id is None:
            raise GravitinoRuntimeException(
                "OSS access key id is not found in the options."
            )

        # get 'oss_secret_access_key' from oss options, if the key is not found, throw an exception
        oss_secret_access_key = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_OSS_SECRET_KEY
        )
        if oss_secret_access_key is None:
            raise GravitinoRuntimeException(
                "OSS secret access key is not found in the options."
            )

        return (
            TIME_WITHOUT_EXPIRATION,
            importlib.import_module("ossfs").OSSFileSystem(
                key=oss_access_key_id,
                secret=oss_secret_access_key,
                endpoint=oss_endpoint,
                **kwargs,
            ),
        )

    def _get_last_modified(self, entry: Dict):
        return entry["LastModified"] if "LastModified" in entry else None

    def _get_actual_prefix(
        self,
        actual_path: str,
    ) -> str:
        ops = infer_storage_options(actual_path)
        if "host" not in ops or "path" not in ops:
            raise GravitinoRuntimeException(
                f"Storage location:{actual_path} doesn't support now."
            )

        return ops["host"] + ops["path"]

    def _get_most_suitable_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, use the
            # secret key credential.
            if isinstance(credential, OSSTokenCredential):
                return credential

        for credential in credentials:
            if isinstance(credential, OSSSecretKeyCredential):
                return credential
        return None


class ABSStorageHandler(StorageHandler):
    """Azure Blob Storage Handler"""

    def storage_type(self):
        return StorageType.ABS

    def get_filesystem(
        self, actual_path: Optional[str] = None, **kwargs
    ) -> AbstractFileSystem:
        return importlib.import_module("adlfs").AzureBlobFileSystem(**kwargs)

    def get_filesystem_with_expiration(
        self,
        credentials: List[Credential],
        catalog_props: Dict[str, str],
        actual_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, AbstractFileSystem]:
        if credentials:
            credential = self._get_most_suitable_credential(credentials)
            if credential is not None:
                expire_time = self._get_expire_time_by_ratio(
                    credential.expire_time_in_ms(),
                    catalog_props,
                )
                if isinstance(credential, ADLSTokenCredential):
                    fs = self.get_filesystem(
                        account_name=credential.account_name(),
                        sas_token=credential.sas_token(),
                        **kwargs,
                    )
                    return expire_time, fs

                if isinstance(credential, AzureAccountKeyCredential):
                    fs = self.get_filesystem(
                        account_name=credential.account_name(),
                        account_key=credential.account_key(),
                        **kwargs,
                    )
                    return expire_time, fs

        # get 'abs_account_name' from abs options, if the key is not found, throw an exception
        abs_account_name = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_AZURE_ACCOUNT_NAME
        )
        if abs_account_name is None:
            raise GravitinoRuntimeException(
                "ABS account name is not found in the options."
            )

        # get 'abs_account_key' from abs options, if the key is not found, throw an exception
        abs_account_key = catalog_props.get(
            GVFSConfig.GVFS_FILESYSTEM_AZURE_ACCOUNT_KEY
        )
        if abs_account_key is None:
            raise GravitinoRuntimeException(
                "ABS account key is not found in the options."
            )

        return (
            TIME_WITHOUT_EXPIRATION,
            self.get_filesystem(
                account_name=abs_account_name,
                account_key=abs_account_key,
                **kwargs,
            ),
        )

    def strip_storage_protocol(self, path: str):
        """Strip the storage protocol from the path.
        We need to remove the protocol and account from the path, for instance,
        the path can be converted from 'abfss://container@account/path' to
        'container/path'.
        :param path: The path
        :return: The stripped path
        """
        ops = infer_storage_options(path)
        return ops["username"] + ops["path"]

    def _get_actual_prefix(
        self,
        actual_path: str,
    ) -> str:
        ops = infer_storage_options(actual_path)
        if "username" not in ops or "host" not in ops or "path" not in ops:
            raise GravitinoRuntimeException(
                f"The location:{actual_path} doesn't support now, the username,"
                f"host and path are required in the  location."
            )
        return f"{self.storage_type().value}://{ops['username']}@{ops['host']}{ops['path']}"

    def _normalize_actual_path_if_needed(
        self, storage_options: Dict[str, Any], actual_path: str
    ) -> str:
        # the actual path may be '{container}/{path}', we need to add the host and username
        # get the path from {container}/{path}
        if not actual_path.startswith(f"{StorageType.ABS}"):
            path_without_username = actual_path[actual_path.index("/") + 1 :]
            return (
                f"{StorageType.ABS.value}://{storage_options['username']}@{storage_options['host']}/"
                f"{path_without_username}"
            )
        return actual_path

    def _get_most_suitable_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, use the
            # account key credential
            if isinstance(credential, ADLSTokenCredential):
                return credential

        for credential in credentials:
            if isinstance(credential, AzureAccountKeyCredential):
                return credential
        return None

    def _get_last_modified(self, entry: Dict):
        return entry["last_modified"]


LOCA_HANDLER = LocalStorageHandler()
HDFS_HANDLER = HDFSStorageHandler()
S3_HANDLER = S3StorageHandler()
GCS_HANDLER = GCSStorageHandler()
OSS_HANDLER = OSSStorageHandler()
ABS_HANDLER = ABSStorageHandler()

storage_prefixes = {
    f"{StorageType.HDFS.value}://": StorageType.HDFS,
    f"{StorageType.LOCAL.value}:/": StorageType.LOCAL,
    f"{StorageType.GCS.value}://": StorageType.GCS,
    f"{StorageType.S3A.value}://": StorageType.S3A,
    f"{StorageType.OSS.value}://": StorageType.OSS,
    f"{StorageType.ABS.value}://": StorageType.ABS,
}

storage_handlers: Dict[StorageType, StorageHandler] = {
    StorageType.LOCAL: LOCA_HANDLER,
    StorageType.HDFS: HDFS_HANDLER,
    StorageType.GCS: GCS_HANDLER,
    StorageType.S3A: S3_HANDLER,
    StorageType.OSS: OSS_HANDLER,
    StorageType.ABS: ABS_HANDLER,
}


def get_storage_handler_by_type(storage_type: StorageType) -> StorageHandler:
    """Get the storage handler by the storage type.
    :param storage_type: The storage type
    :return: The storage handler
    """
    if storage_type not in storage_handlers:
        raise GravitinoRuntimeException(
            f"Storage type: `{storage_type}` doesn't support now."
        )
    return storage_handlers[storage_type]


def get_storage_handler_by_path(actual_path: str) -> StorageHandler:
    """Get the storage handler by the actual path.
    :param actual_path: The actual path
    :return: The storage handler
    """

    for prefix, storage_type in storage_prefixes.items():
        if actual_path.startswith(prefix):
            return get_storage_handler_by_type(storage_type)

    raise GravitinoRuntimeException(
        f"Storage type doesn't support now. Path:{actual_path}"
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.constants.error import ErrorConstants
from gravitino.exceptions.base import (
    AlreadyExistsException,
    IllegalArgumentException,
    MetalakeNotInUseException,
    NoSuchMetalakeException,
    NoSuchTagException,
    NotFoundException,
    TagAlreadyAssociatedException,
    TagAlreadyExistsException,
)
from gravitino.exceptions.handlers.rest_error_handler import RestErrorHandler


class TagErrorHandler(RestErrorHandler):
    """Error handler specific to Tag operations."""

    def handle(self, error_response) -> None:
        error_message = error_response.format_error_message()
        code = error_response.code()
        exception_type = error_response.type()

        if code == ErrorConstants.ILLEGAL_ARGUMENTS_CODE:
            raise IllegalArgumentException(error_message)

        if code == ErrorConstants.NOT_FOUND_CODE:
            if exception_type == NoSuchMetalakeException.__name__:
                raise NoSuchMetalakeException(error_message)

            if exception_type == NoSuchTagException.__name__:
                raise NoSuchTagException(error_message)

            raise NotFoundException(error_message)

        if code == ErrorConstants.ALREADY_EXISTS_CODE:
            if exception_type == TagAlreadyExistsException.__name__:
                raise TagAlreadyExistsException(error_message)

            if exception_type == TagAlreadyAssociatedException.__name__:
                raise TagAlreadyAssociatedException(error_message)

            raise AlreadyExistsException(error_message)

        if code == ErrorConstants.NOT_IN_USE_CODE:
            raise MetalakeNotInUseException(error_message)

        if code == ErrorConstants.INTERNAL_ERROR_CODE:
            raise RuntimeError(error_message)

        super().handle(error_response)


TAG_ERROR_HANDLER = TagErrorHandler()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Optional

from gravitino.api.catalog import Catalog
from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.distributions.distribution import Distribution
from gravitino.api.rel.expressions.sorts.sort_order import SortOrder
from gravitino.api.rel.expressions.transforms.transform import Transform
from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.table import Table
from gravitino.api.rel.table_catalog import TableCatalog
from gravitino.client.base_schema_catalog import BaseSchemaCatalog
from gravitino.client.relational_table import RelationalTable
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.rel.distribution_dto import DistributionDTO
from gravitino.dto.requests.table_create_request import TableCreateRequest
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.table_response import TableResponse
from gravitino.dto.util.dto_converters import DTOConverters
from gravitino.exceptions.handlers.table_error_handler import TABLE_ERROR_HANDLER
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient


class RelationalCatalog(
    BaseSchemaCatalog, TableCatalog
):  # pylint: disable=too-many-ancestors
    """Relational catalog is a catalog implementation

    The `RelationalCatalog` supports relational database like metadata operations,
    for example, schemas and tables list, creation, update and deletion. A Relational
    catalog is under the metalake.
    """

    def __init__(
        self,
        catalog_namespace: Namespace,
        name: str,
        catalog_type: Catalog.Type,
        provider: str,
        audit: AuditDTO,
        rest_client: HTTPClient,
        comment: Optional[str] = None,
        properties: Optional[dict[str, str]] = None,
    ):
        super().__init__(
            catalog_namespace,
            name,
            catalog_type,
            provider,
            comment,
            properties,
            audit,
            rest_client,
        )

    def as_table_catalog(self) -> TableCatalog:
        """Return this relational catalog as a :class:`TableCatalog`.

        This method returns ``self`` to provide access to table-related
        operations defined by the :class:`TableCatalog` interface.

        Returns:
            TableCatalog: The current catalog instance as a ``TableCatalog``.
        """
        return self

    def _check_table_name_identifier(self, identifier: NameIdentifier) -> None:
        """Check whether the `NameIdentifier` of a table is valid.

        Args:
            identifier (NameIdentifier):
                The NameIdentifier to check, which should be "schema.table" format.

        Raises:
            IllegalNameIdentifierException: If the Namespace is not valid.
        """
        NameIdentifier.check(identifier is not None, "NameIdentifier must not be null")
        NameIdentifier.check(
            identifier.name() is not None and identifier.name() != "",
            "NameIdentifier name must not be empty",
        )
        self._check_table_namespace(identifier.namespace())

    def _check_table_namespace(self, namespace: Namespace) -> None:
        """Check whether the namespace of a table is valid, which should be "schema".

        Args:
            namespace (Namespace): The namespace to check.

        Raises:
            IllegalNamespaceException: If the Namespace is not valid.
        """
        Namespace.check(
            namespace is not None and namespace.length() == 1,
            f"Table namespace must be non-null and have 1 level, the input namespace is {namespace}",
        )

    def _get_table_full_namespace(self, table_namespace: Namespace) -> Namespace:
        """Get the full namespace of the table with the given table's short namespace (schema name).

        Args:
            table_namespace (Namespace): The table's short namespace, which is the schema name.

        Returns:
            Namespace: full namespace of the table, which is "metalake.catalog.schema" format.
        """
        return Namespace.of(
            self._catalog_namespace.level(0),
            self._name,
            table_namespace.level(0),
        )

    def _format_table_request_path(self, ns: Namespace) -> str:
        schema_ns = Namespace.of(ns.level(0), ns.level(1))
        return (
            f"{BaseSchemaCatalog.format_schema_request_path(schema_ns)}"
            f"/{encode_string(ns.level(2))}"
            "/tables"
        )

    def create_table(
        self,
        identifier: NameIdentifier,
        columns: list[Column],
        comment: Optional[str] = None,
        properties: Optional[dict[str, str]] = None,
        partitioning: Optional[list[Transform]] = None,
        distribution: Optional[Distribution] = None,
        sort_orders: Optional[list[SortOrder]] = None,
        indexes: Optional[list[Index]] = None,
    ) -> Table:
        self._check_table_name_identifier(identifier)
        req = TableCreateRequest(
            _name=identifier.name(),
            _columns=DTOConverters.to_dtos(columns),
            _comment=comment,
            _properties=properties,
            _sort_orders=DTOConverters.to_dtos(sort_orders),
            _distribution=(
                DTOConverters.to_dto(distribution)
                if distribution is not None
                else DistributionDTO.NONE
            ),
            _partitioning=DTOConverters.to_dtos(partitioning),
            _indexes=DTOConverters.to_dtos(indexes),
        )
        req.validate()
        full_namespace = self._get_table_full_namespace(identifier.namespace())
        resp = self.rest_client.post(
            self._format_table_request_path(full_namespace),
            json=req,
            error_handler=TABLE_ERROR_HANDLER,
        )
        table_resp = TableResponse.from_json(resp.body, infer_missing=True)
        table_resp.validate()
        return RelationalTable(full_namespace, table_resp.table(), self.rest_client)

    def list_tables(self, namespace: Namespace) -> list[NameIdentifier]:
        self._check_table_namespace(namespace)
        full_namespace = self._get_table_full_namespace(namespace)
        resp = self.rest_client.get(
            self._format_table_request_path(full_namespace),
            error_handler=TABLE_ERROR_HANDLER,
        )
        entity_list_resp = EntityListResponse.from_json(resp.body, infer_missing=True)
        entity_list_resp.validate()
        return [
            NameIdentifier.of(ident.namespace().level(2), ident.name())
            for ident in entity_list_resp.identifiers()
        ]

    def load_table(self, identifier: NameIdentifier) -> Table:
        self._check_table_name_identifier(identifier)
        full_namespace = self._get_table_full_namespace(identifier.namespace())
        resp = self.rest_client.get(
            f"{self._format_table_request_path(full_namespace)}"
            f"/{encode_string(identifier.name())}",
            error_handler=TABLE_ERROR_HANDLER,
        )
        table_resp = TableResponse.from_json(resp.body, infer_missing=True)
        table_resp.validate()
        return RelationalTable(full_namespace, table_resp.table(), self.rest_client)

    # TODO: We shall implement the following methods after integration tests for relational table
    def drop_table(self, identifier: NameIdentifier) -> bool:
        raise NotImplementedError("Drop table is not implemented yet.")

    def alter_table(self, identifier: NameIdentifier, *changes) -> Table:
        raise NotImplementedError("Alter table is not implemented yet.")

    def purge_table(self, identifier: NameIdentifier) -> bool:
        raise NotImplementedError("Purge table is not implemented yet.")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from dataclasses import field, dataclass

from dataclasses_json import config

from gravitino.dto.model_version_dto import ModelVersionDTO
from gravitino.dto.responses.base_response import BaseResponse
from gravitino.exceptions.base import IllegalArgumentException


@dataclass
class ModelVersionResponse(BaseResponse):
    """Represents a response for a model version."""

    _model_version: ModelVersionDTO = field(metadata=config(field_name="modelVersion"))

    def model_version(self) -> ModelVersionDTO:
        """Returns the model version."""
        return self._model_version

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException if the model version is not set.
        """
        super().validate()

        if self._model_version is None:
            raise IllegalArgumentException("Model version must not be null")
        if self._model_version.version() is None:
            raise IllegalArgumentException("Model version 'version' must not be null")
        if self._model_version.uris() is None:
            raise IllegalArgumentException("Model version 'uri' must not be null")
        if self._model_version.audit_info() is None:
            raise IllegalArgumentException("Model version 'auditInfo' must not be null")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Dict, List, Optional
from dataclasses import dataclass, field

from dataclasses_json import config

from .job_template_dto import JobTemplateDTO, register_job_template
from ...api.job.job_template import JobType


@register_job_template(JobType.SPARK)
@dataclass
class SparkJobTemplateDTO(JobTemplateDTO):
    """
    Represents a Spark Job Template Data Transfer Object (DTO).

    Note: The class_name field is required for Java/Scala Spark applications but is
    optional for PySpark applications. For PySpark jobs, this field can be None.
    """

    _class_name: Optional[str] = field(
        default=None, metadata=config(field_name="className")
    )
    _jars: Optional[List[str]] = field(default=None, metadata=config(field_name="jars"))
    _files: Optional[List[str]] = field(
        default=None, metadata=config(field_name="files")
    )
    _archives: Optional[List[str]] = field(
        default=None, metadata=config(field_name="archives")
    )
    _configs: Optional[Dict[str, str]] = field(
        default=None, metadata=config(field_name="configs")
    )

    def class_name(self) -> str:
        """Returns the class name of the Spark job."""
        return self._class_name

    def jars(self) -> Optional[List[str]]:
        """Returns the list of JAR files associated with this Spark job template."""
        return self._jars

    def files(self) -> Optional[List[str]]:
        """Returns the list of files associated with this Spark job template."""
        return self._files

    def archives(self) -> Optional[List[str]]:
        """Returns the list of archives associated with this Spark job template."""
        return self._archives

    def configs(self) -> Optional[Dict[str, str]]:
        """Returns the configuration properties for the Spark job."""
        return self._configs



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.function.function_column import FunctionColumn
from gravitino.api.function.function_definition import (
    FunctionDefinition,
    FunctionDefinitions,
)
from gravitino.api.function.function_impl import FunctionImpl
from gravitino.api.function.function_param import FunctionParam
from gravitino.api.rel.types.json_serdes.type_serdes import TypeSerdes
from gravitino.api.rel.types.type import Type
from gravitino.dto.function.function_column_dto import FunctionColumnDTO
from gravitino.dto.function.function_impl_dto import (
    FunctionImplDTO,
    JavaImplDTO,
    PythonImplDTO,
    SQLImplDTO,
    function_impl_dto_from_function_impl,
)
from gravitino.dto.function.function_param_dto import FunctionParamDTO


def _decode_impl(impl_dict: Optional[Dict[str, Any]]) -> Optional[FunctionImplDTO]:
    """Decode a function implementation DTO from a dictionary."""
    if impl_dict is None:
        return None

    language = impl_dict.get("language")
    if language == "SQL":
        return SQLImplDTO.from_dict(impl_dict)
    if language == "JAVA":
        return JavaImplDTO.from_dict(impl_dict)
    if language == "PYTHON":
        return PythonImplDTO.from_dict(impl_dict)

    raise ValueError(f"Unsupported implementation language: {language}")


def _decode_impls(impls_list: List[Dict[str, Any]]) -> List[FunctionImplDTO]:
    """Decode a list of function implementation DTOs from a list of dictionaries."""
    if impls_list is None:
        return []
    return [_decode_impl(impl) for impl in impls_list]


def _encode_impl(impl: Optional[FunctionImplDTO]) -> Optional[Dict[str, Any]]:
    """Encode a function implementation DTO to a dictionary."""
    if impl is None:
        return None
    result = impl.to_dict()
    result["language"] = impl.language().name
    return result


def _encode_impls(impls: List[FunctionImplDTO]) -> List[Dict[str, Any]]:
    """Encode a list of function implementation DTOs to a list of dictionaries."""
    if impls is None:
        return []
    return [_encode_impl(impl) for impl in impls]


@dataclass
class FunctionDefinitionDTO(FunctionDefinition, DataClassJsonMixin):
    """DTO for function definition."""

    _parameters: Optional[List[FunctionParamDTO]] = field(
        default=None, metadata=config(field_name="parameters")
    )
    _return_type: Optional[Type] = field(
        default=None,
        metadata=config(
            field_name="returnType",
            encoder=TypeSerdes.serialize,
            decoder=TypeSerdes.deserialize,
        ),
    )
    _return_columns: Optional[List[FunctionColumnDTO]] = field(
        default=None, metadata=config(field_name="returnColumns")
    )
    _impls: Optional[List[FunctionImplDTO]] = field(
        default=None,
        metadata=config(
            field_name="impls",
            encoder=_encode_impls,
            decoder=_decode_impls,
        ),
    )

    def parameters(self) -> List[FunctionParam]:
        """Returns the parameters for this definition."""
        return list(self._parameters) if self._parameters else []

    def return_type(self) -> Optional[Type]:
        """Returns the return type for scalar or aggregate functions."""
        return self._return_type

    def return_columns(self) -> List[FunctionColumn]:
        """Returns the output columns for a table-valued function."""
        if self._return_columns is None:
            return FunctionDefinition.EMPTY_COLUMNS
        return [col.to_function_column() for col in self._return_columns]

    def impls(self) -> List[FunctionImpl]:
        """Returns the implementations associated with this definition."""
        if self._impls is None:
            return []
        return [impl.to_function_impl() for impl in self._impls]

    def to_function_definition(self) -> FunctionDefinition:
        """Convert this DTO to a FunctionDefinition instance."""
        params = (
            [p.to_function_param() for p in self._parameters]
            if self._parameters
            else []
        )
        impls = [impl.to_function_impl() for impl in self._impls] if self._impls else []

        if self._return_type is not None:
            return FunctionDefinitions.of(params, self._return_type, impls)
        if self._return_columns and len(self._return_columns) > 0:
            cols = [col.to_function_column() for col in self._return_columns]
            return FunctionDefinitions.of_table(params, cols, impls)
        # Fallback for backward compatibility
        return FunctionDefinitions.SimpleFunctionDefinition(params, None, None, impls)

    @classmethod
    def from_function_definition(
        cls, definition: FunctionDefinition
    ) -> "FunctionDefinitionDTO":
        """Create a FunctionDefinitionDTO from a FunctionDefinition instance."""
        param_dtos = (
            [
                (
                    p
                    if isinstance(p, FunctionParamDTO)
                    else FunctionParamDTO.from_function_param(p)
                )
                for p in definition.parameters()
            ]
            if definition.parameters()
            else []
        )

        return_column_dtos = None
        if definition.return_columns() and len(definition.return_columns()) > 0:
            return_column_dtos = [
                FunctionColumnDTO.from_function_column(col)
                for col in definition.return_columns()
            ]

        impl_dtos = (
            [function_impl_dto_from_function_impl(impl) for impl in definition.impls()]
            if definition.impls()
            else []
        )

        return cls(
            _parameters=param_dtos,
            _return_type=definition.return_type(),
            _return_columns=return_column_dtos,
            _impls=impl_dtos,
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, FunctionDefinitionDTO):
            return False
        return (
            self._parameters == other._parameters
            and self._return_type == other._return_type
            and self._return_columns == other._return_columns
            and self._impls == other._impls
        )

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self._parameters) if self._parameters else None,
                self._return_type,
                tuple(self._return_columns) if self._return_columns else None,
                tuple(self._impls) if self._impls else None,
            )
        )
